#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Graphical text editing functionalities:
        - Autocompletion
        - Syntax highlighting
        - Automatic placement

    Copyright (c) 2012-2020 European Space Agency

    Designed and implemented by Maxime Perrotin for the TASTE project

    Contact: maxime.perrotin@esa.int
"""

import string
import logging

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *
from . import undoCommands

__all__ = ['EditableText']

LOG = logging.getLogger(__name__)

# pylint: disable=R0904
class Completer(QGraphicsProxyWidget):
    ''' Class for handling text autocompletion in the SDL scene '''
    def __init__(self, parent):
        ''' Create an autocompletion list popup '''
        widget = QListWidget()
        super().__init__(parent)
        self.setWidget(widget)
        self.string_list = QStringListModel()
        self._completer = QCompleter()
        self.parent = parent
        self._completer.setCaseSensitivity(Qt.CaseInsensitive)
        # For some reason the default minimum size is (61,61)
        # Set it to 0 so that the size of the box is not taken
        # into account when it is hidden.
        self.setMinimumSize(0, 0)
        self.prepareGeometryChange()
        self.resize(0, 0)
        self.hide()

    def set_completer_list(self):
        ''' Set list of items for the autocompleter popup '''
        compl = [item.replace('-', '_') for item in
                 self.parent.parentItem().completion_list]
        self.string_list.setStringList(compl)
        self._completer.setModel(self.string_list)

    def set_completion_prefix(self, completion_prefix):
        '''
            Set the current completion prefix (user-entered text)
            and set the corresponding list of words in the popup widget
        '''
        self._completer.setCompletionPrefix(completion_prefix)
        self.widget().clear()
        count = self._completer.completionCount()
        for i in range(count):
            self._completer.setCurrentRow(i)
            self.widget().addItem(self._completer.currentCompletion())
        self.prepareGeometryChange()
        if count:
            self.resize(self.widget().sizeHintForColumn(0) + 40, 70)
        else:
            self.resize(0, 0)
        return count

    # pylint: disable=C0103
    def keyPressEvent(self, e):
        super().keyPressEvent(e)
        if e.key() == Qt.Key_Escape:
            self.parentItem().setFocus()
        # Consume the event so that it is not repeated at EditableText level
        e.accept()

    # pylint: disable=C0103
    def focusOutEvent(self, event):
        ''' When the user leaves the popup, return focus to parent '''
        super().focusOutEvent(event)
        self.hide()
        self.resize(0, 0)
        self.parentItem().setFocus()


# pylint: disable=R0904
class Highlighter(QSyntaxHighlighter):
    ''' Class for handling syntax highlighting in editable text '''
    def __init__(self, parent, blackbold_patterns, redbold_patterns):
        ''' Define highlighting rules - inputs = lists of patterns '''
        super().__init__(parent)
        self.highlighting_rules = []

        # Black bold items (allowed keywords)
        black_bold_format = QTextCharFormat()
        black_bold_format.setFontWeight(QFont.Bold)
        for pattern in blackbold_patterns:
            self.highlighting_rules.append(
                    (QRegularExpression(pattern,
                                        QRegularExpression.CaseInsensitiveOption),
                                        black_bold_format))

        # Red bold items (reserved keywords)
        red_bold_format = QTextCharFormat()
        red_bold_format.setFontWeight(QFont.Bold)
        red_bold_format.setForeground(Qt.red)

        for pattern in redbold_patterns:
            self.highlighting_rules.append(
                    (QRegularExpression(pattern,
                                        QRegularExpression.CaseInsensitiveOption),
                                        red_bold_format))

        # Comments
        comment_format = QTextCharFormat()
        comment_format.setForeground(Qt.darkBlue)
        comment_format.setFontItalic(True)
        self.highlighting_rules.append(
                (QRegularExpression('--[^\n]*'), comment_format))

    # pylint: disable=C0103
    def highlightBlock(self, text):
        ''' Redefined function to apply the highlighting rules '''
        for expression, formatter in self.highlighting_rules:
            index = expression.globalMatch(text)
            while (index.hasNext()):
                match = index.next()
                length = match.capturedLength()
                self.setFormat(match.capturedStart(), length, formatter)


def zValueItems(item):
    ''' Find all items that require a zValue increase to ensure that both
    the text and completer are visible. '''
    if item.parent is not None:
        yield from zValueItems(item.parent)
    yield item

# pylint: disable=R0902
class EditableText(QGraphicsTextItem):
    '''
        Editable text area inside symbols
        Includes autocompletion when parent item needs it
    '''
    default_cursor = Qt.IBeamCursor
    hasParent = False
    word_under_cursor = Signal(str)

    def __init__(self, parent, text='...', hyperlink=None):
        super().__init__(parent)
        self.parent = parent
        self.setFont(QFont('Ubuntu', 10))
        self.completer = Completer(self)
        self.completer.widget().itemActivated.connect(self.completion_selected)
        self.hyperlink = hyperlink
        self.setOpenExternalLinks(True)
        if hyperlink:
            self.setHtml('<a href="{hlink}">{text}</a>'.format
                    (hlink=hyperlink, text=text.replace('\n', '<br>')))
        else:
            self.setPlainText(text)
        self.setTextInteractionFlags( Qt.TextSelectableByMouse
                                     | Qt.TextEditable
                                     | Qt.TextSelectableByKeyboard
                                     | Qt.LinksAccessibleByMouse
                                     | Qt.LinksAccessibleByKeyboard)
        self.completer_has_focus = False
        self.editing = False
        self.try_resize()
        self.highlighter = Highlighter(
                self.document(), parent.blackbold, parent.redbold)
        self.completion_prefix = ''
        self.old_word = ''  # needed to detect change of word under cursor
        #self.set_textbox_position()
        self.set_text_alignment()
        # Increase the Z value of the text area so that the autocompleter
        # always appear on top of text's siblings (parents's followers)
        self.setZValue(self.zValue() + 1)
        # context is used for advanced autocompletion
        self.context = ''
        # Set cursor when mouse goes over the text
        self.setCursor(self.default_cursor)
        # Activate cache mode to boost rendering by calling paint less often
        # Removed - does not render text properly (eats up the right part)
        # self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
        self.force_focus = False

    def set_text_alignment(self):
        ''' Apply the required text alignment within the text box '''
        alignment = self.parent.text_alignment
        self.setTextWidth(self.boundingRect().width())
        fmt = QTextBlockFormat()
        fmt.setAlignment(alignment)
        cursor = self.textCursor()
        cursor.select(QTextCursor.Document)
        cursor.mergeBlockFormat(fmt)
        cursor.clearSelection()
        self.setTextCursor(cursor)

    def set_textbox_position(self):
        ''' Compute the textbox position '''
        parent_rect = self.parent.boundingRect()
        rect = self.boundingRect()
        # Use parent symbol alignment requirement
        # Does not support right alignment (just add it when needed)
        alignment = self.parent.textbox_alignment
        rect_center = parent_rect.center() - rect.center()
        if alignment & Qt.AlignLeft:
            x_pos = 0
        elif alignment & Qt.AlignHCenter:
            x_pos = rect_center.x()
        else:
            x_pos = 0
        if alignment & Qt.AlignTop:
            y_pos = 0
        elif alignment & Qt.AlignVCenter:
            y_pos = rect_center.y()
        elif alignment & Qt.AlignBottom:
            y_pos = parent_rect.height()
        else:
            y_pos = 0
        self.setPos(x_pos, y_pos)

    def try_resize(self):
        '''
            Request a resizing of the parent item (symbol) when text changes
        '''
        minW, minH = self.parent.min_size
        self.setTextWidth(-1)
        parent_rect = self.parent.boundingRect()
        rect = self.boundingRect()
        #if rect.width() + 15 > parent_rect.width():
        parent_rect.setWidth(max(rect.width() + 15, minW))
        parent_rect.setHeight(max(rect.height(), minH))
        self.parent.resize_item(parent_rect)
        self.set_textbox_position()

    @Slot(QListWidgetItem)
    def completion_selected(self, item):
        '''
            Slot connected to the autocompletion popup,
            invoked when selection is made
        '''
        if not(self.textInteractionFlags() & Qt.TextEditable):
            self.completer.hide()
            return
        text_cursor = self.textCursor()
        # Go back to the previously saved cursor position
        text_cursor.setPosition(self.cursor_position)
        extra = len(item.text()) - len(self.completion_prefix)
        if extra > 0:
            if len(self.completion_prefix):
                # Move back left only if there is a word to replace
                text_cursor.movePosition(QTextCursor.Left)
                text_cursor.movePosition(QTextCursor.EndOfWord)
            text_cursor.insertText(item.text()[-extra:])
            self.setTextCursor(text_cursor)
        self.completer_has_focus = False
        self.completer.hide()
        self.try_resize()


    def context_completion_list(self, force=False):
        ''' Advanced context-dependent autocompletion '''
        # Select text from the beginning of a line to the cursor position
        # Then keep the last word including separators ('!' and '.')
        # This word (e.g. variable!field!subfield) is then used to update
        # the autocompletion list.
        cursor = self.textCursor()
        pos = cursor.positionInBlock() - 1
        cursor.select(QTextCursor.BlockUnderCursor)
        context = self.context
        try:
            # If not the first line of the text, Qt adds u+2029 as 1st char
            line = cursor.selectedText().replace(u'\u2029', '')
            if line[pos] in string.ascii_letters + '!' + '.' + '_' + ':':
                self.context = line[slice(0, pos + 1)].split()[-1]
            else:
                self.context = ''
        except IndexError:
            pass
        if (context != self.context) or force:
            self.completer.set_completer_list()


    # pylint: disable=C0103
    def keyPressEvent(self, event):
        '''
            Activate the autocompletion window if relevant
        '''
        super().keyPressEvent(event)
        # Typing Esc allows to stop editing text:
        if event.key() == Qt.Key_Escape:
            self.clearFocus()
            return
        # When completer is displayed, give it the focus with down key
        if self.completer.isVisible() and event.key() == Qt.Key_Down:
            self.completer_has_focus = True
            self.completer.setFocusProxy(None)
            self.completer.widget().setFocus()
            return
        self.try_resize()
        text_cursor = self.textCursor()
        text_cursor.select(QTextCursor.WordUnderCursor)
        self.completion_prefix = text_cursor.selectedText()

        # "self.completion_prefix" is the complete word under the cursor
        if self.completion_prefix != self.old_word:
            self.word_under_cursor.emit(self.completion_prefix)
            self.old_word = self.completion_prefix

        self.context_completion_list(force=(event.key()==Qt.Key_F8))

        completion_count = self.completer.set_completion_prefix(
                self.completion_prefix)
        if(completion_count > 0 and len(self.completion_prefix) > 1) or(
                event.key() == Qt.Key_F8):
            # Save the position of the cursor
            self.cursor_position = self.textCursor().position()
            # Computing the coordinates of the completer
            # No direct Qt function for that.. doing it the hard way
            pos = self.textCursor().positionInBlock()
            block = self.textCursor().block()
            layout = block.layout()
            line = layout.lineForTextPosition(pos)
            rect = line.rect()
            relative_x, _ = line.cursorToX(pos)
            layout_pos = layout.position()
            pos_x = relative_x + layout_pos.x()
            pos_y = rect.y() + rect.height() + layout_pos.y()

            self.completer.setPos(pos_x, pos_y)
            self.completer.show()
            self.completer.ensureVisible()
            self.force_focus = True   # avoid unwanted syntax checks
            # Make sure parent item has higher visibility than its siblings
            # (useful in decision branches)
            self.parent.setZValue(1)
            self.completer.setFocusProxy(self)
            self.setTabChangesFocus(True)
        else:
            self.completer.setFocusProxy(None)
            self.completer.hide()
            self.completer.resize(0, 0)
            self.setFocus()
        self.completer_has_focus = False

    def mousePressEvent(self, event):
        '''
            If the completer box is active while the user clicks on another
            area of the text box, make it disappear first
        '''
        if self.completer.isVisible():
            self.completer.hide()
            self.completer.resize(0, 0)
        super().mousePressEvent(event)


    # pylint: disable=C0103
    def focusOutEvent(self, event):
        '''
            When the user stops editing, this function is called
            In that case, hide the completer if it is not the item
            that got the focus.
        '''
        if not self.editing:
            return super().focusOutEvent(event)
        if self.completer and not self.completer_has_focus:
            self.completer.hide()
            self.completer.resize(0, 0)
        if self.force_focus:
            # when user double-clicks on the Completer, it may be out of
            # the editable text. It is not right to leave the focus in that
            # case, as this would generate a syntax check while in fact
            # user is not done editing text
            self.setFocus()
            self.force_focus = False
            return
        if not self.completer or not self.completer.isVisible():
            # Trigger a select - side effect makes the toolbar update
            try:
                self.parent.select(True)
            except AttributeError:
                # Some parents may not be selectable (e.g. Signalroute)
                pass
            self.editing = False
            text_cursor = self.textCursor()
            if text_cursor.hasSelection():
                text_cursor.clearSelection()
                self.setTextCursor(text_cursor)
            # If something has changed, check syntax and create undo command
            if(self.oldSize != self.parent.boundingRect()
                    or self.parent.syntax_error or self.oldText != str(self)):
                # Call syntax checker from item containing the text (if any)
                if self.scene().check_syntax(self.parent):
                    # Keep focus
                    self.setFocus()
                    return
                # Update class completion list
                self.scene().update_completion_list(self.parentItem())
                # Create undo command, including possible CAM
                with undoCommands.UndoMacro(self.scene().undo_stack, 'Text'):
                    undo_cmd = undoCommands.ResizeSymbol(
                                          self.parent, self.oldSize,
                                          self.parent.boundingRect())
                    self.scene().undo_stack.push(undo_cmd)
                    try:
                        self.parent.cam(self.parent.pos(),
                                        self.parent.pos())
                    except AttributeError:
                        # Some parents may not have CAM function (e.g. Channel)
                        pass

                    undo_cmd = undoCommands.ReplaceText(self, self.oldText,
                                                        str(self))
                    self.scene().undo_stack.push(undo_cmd)
        self.set_text_alignment()
        # Reset Z-Values that were increased when getting focus
        for each in zValueItems(self.parent):
            each.setZValue(each.zValue() - 1)
        super().focusOutEvent(event)

    # pylint: disable=C0103
    def focusInEvent(self, event):
        ''' When user starts editing text, save previous state for Undo '''
        super().focusInEvent(event)
        # Change the Z-value of items to make sure the
        # completer is always be on top of other symbols
        for each in zValueItems(self.parent):
            each.setZValue(each.zValue() + 1)

        # Trigger a select - side effect makes the toolbar update
        try:
            self.parent.select(True)
        except AttributeError:
            # Some parents may not be selectable
            pass
        # Update completer list of keywords
        self.context = ''
        self.completer.set_completer_list()
        # Clear selection otherwise the "Delete" key may delete other items
        self.scene().clearSelection()
        # Set width to auto-expand, and disables alignment, while editing:
        self.setTextWidth(-1)
        if not self.editing:
            self.oldText = str(self)
            self.oldSize = self.parent.boundingRect()
            self.editing = True

    def __str__(self):
        ''' Print the text inside the symbol '''
        return self.toPlainText()



"""
                                   Q G E N                                    
                                                                              
                       Copyright (C) 2011-2018, AdaCore                       
                       Copyright (C) 2011-2018, IB Krates                     
                                                                              
   This is free software;  you can redistribute it  and/or modify it  under   
   terms of the  GNU General Public License as published  by the Free Soft-   
   ware  Foundation;  either version 3,  or (at your option) any later ver-   
   sion.  This software is distributed in the hope  that it will be useful,   
   but WITHOUT ANY WARRANTY;  without even the implied warranty of MERCHAN-   
   TABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public   
   License for  more details.  You should have  received  a copy of the GNU   
   General  Public  License  distributed  with  this  software;   see  file   
   COPYING3.  If not, go to http://www.gnu.org/licenses for a complete copy   
   of the license.                                                            
                                                                              
"""
import os
import subprocess
import shutil
import logging

LOG = logging.getLogger(__name__)

def call_qgensdl(options):

    ''' Call QGen-SDL tool with the required arguments '''

    lang=''
    if options.toC:
        lang = 'c'
    elif options.toAda:
        lang = 'ada'

    LOG.debug('Generating ' + lang + ' code using QGen from ' + str(options.files))

    outfolder = 'qgen_generated_' + lang
    cmd = ["qgen-sdl", '--language', lang,
            '--output', outfolder,
             '--type-prefix', 'asn1Scc',
             options.files[0]]

    if os.path.exists(outfolder):
            shutil.rmtree(outfolder, ignore_errors=True)

    LOG.debug('QGen-sdl command: ' + str(cmd))
    try:
        p1 = subprocess.Popen(cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
        errcode = p1.wait()
        stdout, stderr = p1.communicate()
        if errcode:
            LOG.error(stderr)
        LOG.info(stdout)
    except OSError as e:
        errcode = 1
        if e.errno == os.errno.ENOENT:
            LOG.error ('QGen-SDL tool is not found on your path')
        else:
            raise

    return (errcode)


#!/usr/bin/env python3

"""
    OpenGEODE - A tiny, free SDL Editor for TASTE

    SDL is the Specification and Description Language (Z100 standard from ITU)

    Copyright (c) 2012-2022 Maxime Perrotin
    Copyright (c) 2012-2022 European Space Agency

    Designed and implemented by Maxime Perrotin

    This module is responsible for the rendering of statecharts using
    the graphviz library.

    Credits:
    Rendering algorithm to transform graphviz b-splines to to Qt bezier curves
    was developed by Steve Dodier-Lazaro (www.mupuf.org)

    Contact: maxime.perrotin@esa.int
"""

import os
import logging
import traceback
from collections import defaultdict
from functools import partial
from itertools import chain
import re

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *
from PySide6.QtUiTools import QUiLoader

# import resource file to get the configuration widget
from . import icons

g_statechart_lock = False

try:
    import pygraphviz as dotgraph
except ImportError:
    print('pygraphviz not found. Statechart module will not work\n'
          'If you are using Debian (or Ubuntu) you can install it from\n'
          'the official repos: sudo apt-get install python3-pygraphviz')

from . import genericSymbols
from .Connectors import Edge

RENDER_DPI = {'X': 72, 'Y': 72}   # default to 72, in case there is no view
G_SYMBOLS = set()
EDGES = []

LOG = logging.getLogger(__name__)


# pylint: disable=R0904
class Record(genericSymbols.HorizontalSymbol):
    ''' Graphviz node - it is floating and has no parent'''
    _unique_followers = []
    _insertable_followers = ['Record', 'Diamond', 'Stop']
    _terminal_followers = []
    textbox_alignment = (Qt.AlignTop | Qt.AlignHCenter)

    # Minimum size for symbol
    min_size = (70, 35)

    def __init__(self, node, graph):
        ''' Initialization: compute the polygon shape '''
        self.name = node['name']
        # Keep the full path to the state (with parent state names)
        # This is needed for symbol highlightning of parallel states
        self.path = node['path']
        super().__init__(x=node['pos'][0],
                y=node['pos'][1], text=self.name)
        self.set_shape(node['width'], node['height'])
        self.setBrush(QBrush(QColor(255, 255, 202)))
        self.graph = graph
        if 'properties' in node:
            property_box = QGraphicsTextItem(self)
            property_box.setPos(5, 30)
            property_box.setPlainText(node['properties'])
        # Text in statecharts is read-only:
        self.text.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.text.set_textbox_position()

    def set_shape(self, width, height):
        ''' Define the polygon shape from width and height '''
        path = QPainterPath()
        path.moveTo(0, 25)
        path.lineTo(width, 25)
        path.addRoundedRect(0, 0, width, height, 15, 15)
        self.setPath(path)
        LOG.debug ("Set shape of state " + str(self))
        #LOG.debug(traceback.print_stack())
        super().set_shape(width, height)

    def resize_item(self, _):
        ''' Redefine the resizing function - forbid resizing '''
        pass

    def __str__(self):
        ''' User-friendly information about the node '''
        # must return the state name only, as it is used in the renderer
        # in V2, there was a separate __unicode__ function
        return self.name

    def mouse_release(self, _):
        ''' After moving item, ask dot to recompute the edges '''
        # Discard mouse release default action (CAM, UndoCommand)
        pass

    def mouse_move(self, event):
        ''' Disallow moving the symbols - this scene is auto-generated '''
        pass


# pylint: disable=R0904
class Point(genericSymbols.HorizontalSymbol):
    ''' Graphviz point node - used for START transition'''
    _unique_followers = []
    _insertable_followers = ['Record', 'Diamond', 'Stop']
    _terminal_followers = []
    textbox_alignment = (Qt.AlignBottom | Qt.AlignHCenter)
    has_text_area = True # False (in nested state, START can have a label)

    def __init__(self, node, graph):
        ''' Initialization: compute the polygon shape '''
        self.name = node['name']
        if len (self.name) > 5:
            #  remove the _START suffix in nested state named label
            label = self.name[0:-6]
        else:
            label=''
        super().__init__(x=node['pos'][0], y=node['pos'][1],
                text=label)
        self.set_shape(node['width'], node['height'])
        self.setBrush(QBrush(Qt.black))
        self.graph = graph
        # Text is read only
        self.text.setTextInteractionFlags(Qt.TextBrowserInteraction)

    def set_shape(self, width, height):
        ''' Define the polygon shape from width and height '''
        path = QPainterPath()
        path.addEllipse(0, 0, width, height)
        self.setPath(path)
        super().set_shape(width, height)

    def resize_item(self, _):
        ''' Redefine the resizing function - forbid resizing '''
        pass

    def __str__(self):
        ''' User-friendly information about the node '''
        return('Point ' + self.name + ' at pos ' + str(self.pos()) +
                ' bounding rect = ' + str(self.boundingRect()))

    def mouse_release(self, _):
        ''' After moving item, ask dot to recompute the edges '''
        #self.scene().refresh()
        update(self.scene())

    def mouse_move(self, event):
        ''' Disallow moving the symbols - this scene is auto-generated '''
        pass


# pylint: disable=R0904
class Diamond(genericSymbols.HorizontalSymbol):
    ''' Graphviz node - it is floating and has no parent'''
    _unique_followers = []
    _insertable_followers = ['Record', 'Stop']
    _terminal_followers = []
    textbox_alignment = (Qt.AlignTop
                         | Qt.AlignHCenter)
    has_text_area = False

    def __init__(self, node, graph):
        ''' Initialization: compute the polygon shape '''
        self.name = node['name']
        super().__init__(x=node['pos'][0], y=node['pos'][1])
        self.set_shape(node['width'], node['height'])
        self.graph = graph

    def set_shape(self, width, height):
        ''' Define the polygon shape from width and height '''
        path = QPainterPath()
        path.moveTo(width / 2, 0)
        path.lineTo(width, height / 2)
        path.lineTo(width / 2, height)
        path.lineTo(0, height / 2)
        path.lineTo(width / 2, 0)
        self.setPath(path)
        super().set_shape(width, height)

    def resize_item(self, _):
        ''' Redefine the resizing function - forbid resizing '''
        pass

    def __str__(self):
        ''' User-friendly information about the node '''
        return('Node ' + self.name + ' at pos ' + str(self.pos()) +
                ' bounding rect = ' + str(self.boundingRect()))

    def mouse_release(self, _):
        ''' After moving item, ask dot to recompute the edges '''
        #update(self.scene())
        pass

    def mouse_move(self, event):
        ''' Disallow moving the symbols - this scene is auto-generated '''
        pass


# pylint: disable=R0904
class Stop(genericSymbols.HorizontalSymbol):
    ''' Graphviz state exit node - it is floating and has no parent'''
    _unique_followers = []
    _insertable_followers = []
    _terminal_followers = []
    textbox_alignment = (Qt.AlignTop
                         | Qt.AlignHCenter)
    has_text_area = True

    def __init__(self, node, graph):
        ''' Initialization: compute the polygon shape '''
        self.name = node['name']
        super().__init__(x=node['pos'][0], y=node['pos'][1],
                                   text=self.name)
        self.set_shape(node['width'], node['height'])
        self.graph = graph
        # Text in statecharts is read-only:
        self.text.setTextInteractionFlags(Qt.TextBrowserInteraction)

    def set_shape(self, width, height):
        ''' Define the polygon shape from width and height '''
        width, height = width, height
        path = QPainterPath()
        path.lineTo(width, height)
        path.moveTo(width, 0)
        path.lineTo(0, height)
        self.setPath(path)
        super().set_shape(width, height)

    def resize_item(self, _):
        ''' Redefine the resizing function - forbid resizing '''
        pass

    def __str__(self):
        ''' User-friendly information about the node '''
        return('Stop node ' + self.name + ' at pos ' + str(self.pos()) +
                ' bounding rect = ' + str(self.boundingRect()))

    def mouse_release(self, _):
        ''' After moving item, ask dot to recompute the edges '''
        pass

    def mouse_move(self, event):
        ''' Disallow moving the symbols - this scene is auto-generated '''
        pass


def edges(scene, node):
    ''' Return all edges of a given node '''
    for item in scene.items():
        if isinstance(item, Edge) and node in (
                item.edge['source'], item.edge['target']):
            yield item


# Front end part - use Graphviz to parse the input file and transform it
# pylint: disable=R0914
def preprocess_edges(my_graph, nodes, bounding_rect, dpi):
    ''' Parse edges and put them in a clean datastructure '''
    edges = []
    bb_height = bounding_rect[3]
    for edge in my_graph.edges_iter():
        new_edge = {}
        source, target = edge
        for node in nodes:
            if node.name == source:
                new_edge['source'] = node
            if node.name == target:
                new_edge['target'] = node
        new_edge['label'] = edge.attr.get('label')
        try:
            new_edge['lp'] = [float(val) for val in edge.attr['lp'].split(',')]
        except (ValueError, AttributeError):
            pass
        else:
            # translate the label position from dot-coordinates to Qt
            # Note: lp is the position of the CENTER of the label
            new_edge['lp'][0] *= RENDER_DPI['X'] / dpi
            new_edge['lp'][1] = (
                    bb_height - new_edge['lp'][1]) * (RENDER_DPI['Y'] / dpi)

        pos = edge.attr['pos'].split()

        if pos[0].startswith('e'):
            new_edge['end'] = [float(val)
                    for val in pos.pop(0).split(',')[1:]]
            # And translate position
            new_edge['end'][0] *= (RENDER_DPI['X'] / dpi)
            new_edge['end'][1] = (
                    bb_height - new_edge['end'][1]) * (RENDER_DPI['Y'] / dpi)

        elif pos[0].startswith('s'):
            new_edge['start'] = [float(val)
                    for val in pos.pop(0).split(',')[1:]]
            # And translate position
            new_edge['start'][0] *= (RENDER_DPI['X'] / dpi)
            new_edge['start'][1] = (
                    bb_height - new_edge['start'][1]) * (RENDER_DPI['Y'] / dpi)

        spline = []
        for entry in pos:
            point = [float(val) for val in entry.split(',')]
            point[0] *= (RENDER_DPI['X'] / dpi)
            point[1] = (bb_height - point[1]) * (RENDER_DPI['Y'] / dpi)

            spline.append(point)
        new_edge['spline'] = spline
        edges.append(new_edge)
    return edges


def preprocess_nodes(my_graph, bounding_rect, dpi):
    ''' Parse the nodes from the graph and create a clean datastructure '''
    # Graphviz calculates coordinates starting on the bottom-left corner,
    # while Qt uses the top-left corner, so we must translate all y-coord.
    bb_height = bounding_rect[3]
    nodes = []
    attrs = []
    for each in my_graph.subgraphs():
        # Transform cluster nodes into nodes (attributes differ)
        llx, lly, urx, ury = each.graph_attr['bb'].split(',')
        wid = float(urx) - float(llx)
        hei = float(ury) - float(lly)
        attr = {}
        attr.update(pos='{0},{1}'.format(llx, lly),
                         width=wid,
                         height=hei,
                         shape=each.graph_attr['shape'],
                         label=each.graph_attr['label'],
                         style=each.graph_attr['style'],
                         name='cluster_' + each.graph_attr['label'],
                         kind='cluster')
        attrs.append(attr)
    for node in my_graph.nodes_iter():
        if not dict(node.attr):
            continue
        node.attr.update(name=node.name)
        attrs.append(node.attr)
    for node in attrs:
        new_node = {}
        # Get the node shape attribute - default is record
        new_node['shape'] = node.get('shape') or 'record'
        # node main name
        new_node['name'] = node['name']
        # node full path (for nested/parallel states)
        new_node['path'] = node['comment']
        try:
            # node complete label (can contain several compartments)
            properties = (
                    node['label'][1:-2].split('|')[1].replace('\\', '\n'))
            new_node['properties'] = properties
        except IndexError:
            pass
        # transform width and height from inches to pixels
        if node.get('kind', '') == 'cluster':
            new_node['width'] = node['width']
            new_node['height'] = node['height']
        else:
            new_node['width'] = float(node['width']) * RENDER_DPI['X']
            new_node['height'] = float(node['height']) * RENDER_DPI['Y']
        # get the position of the CENTER of the node
        center_pos = [float(val) for val in node['pos'].split(',')]
        # apply dpi-conversion from 72 to actual screen DPI
        center_pos[0] *= (RENDER_DPI['X'] / dpi)

        # translate y-coord from bottom-left to top-left corner
        center_pos[1] = (bb_height - center_pos[1]) * (RENDER_DPI['Y'] / dpi)
        new_node['pos'] = [center_pos[0] - (new_node['width'] / 2.0),
                           center_pos[1] - (new_node['height'] / 2.0)]
        nodes.append(new_node)
    return nodes


def update(scene):
    '''
        Parse the graph symbols and create a graphviz graph
        Used to update the edges in case user moves nodes.
        This function is disabled because move of symbols is not possible
        anymore.
    '''
    return
    nodes = [{'name':node.name, 'pos':
        node.mapToScene(node.boundingRect().center()),
        'shape': type(node), 'width': node.boundingRect().width(),
        'height': node.boundingRect().height()}
        for node in scene.visible_symb if isinstance(node,
                                               (Point, Diamond, Record, Stop))]
    graph = dotgraph.AGraph(
            strict=False, directed=True, splines='spline', start='rand')

    lookup = {Point: 'point', Record: 'record',
              Diamond: 'diamond', Stop: 'square'}
    for node in nodes:
        center_pos = node['pos']
        bb_height = scene.itemsBoundingRect().height()

        dpi = 72.0

        # Convert to graphviz coordinates / adapt DPI
        center_x = center_pos.x() * (dpi / RENDER_DPI['X'])
        center_y = (bb_height - center_pos.y()) * (dpi / RENDER_DPI['Y'])

        pos = str('{x},{y}'.format(x=float(center_x), y=float(center_y)))

        if node['shape'] in (Point, Diamond, Stop):
            graph.add_node(node['name'], pos=pos, shape=lookup[node['shape']],
                fixedsize='true', width=node['width'] / RENDER_DPI['X'],
                height=node['height'] / RENDER_DPI['Y'], pin=True)
        else:
            graph.add_node(node['name'], pos=pos, pin=True,
                           shape=lookup[node['shape']])

    # Keep edges from the previous graph
    for edge in EDGES:
        graph.add_edge(edge, label=edge.attr.get('label') or '')

    #print graph.to_string()
    if nodes:
        before = scene.itemsBoundingRect().center()
        #before_pos = graph.get_node(nodes[0]['name']).attr['pos']
        render_statechart(scene, {'graph': graph, 'children': {}},
                          keep_pos=True)
        #after_pos = graph.get_node(nodes[0]['name']).attr['pos']
        #print before_pos,after_pos
        delta = scene.itemsBoundingRect().center() - before
        # graphviz translates the graph to pos (0, 0) -> move it back
        # to the exact graphical position where the user clicked
        for item in scene.visible_symb:
            item.position = QPointF(item.position - delta)


def render_statechart(scene, graphtree=None, keep_pos=False, dump_gfx=''):
    ''' Render a graphviz/dot statechart on the QGraphicsScene
        set a filename to "dump_gfx" parameter to create a PNG of the graph
        input is resulting from sdl_to_statechart, it contains a tree of graphs
        in case of composite states.
    '''
    if graphtree == None:
        raise TypeError ("Statechart.py: 'render_statechart'"
                         " called with no graph tree")
    # dot uses a 72 dpi value for converting its position coordinates
    # Get actual rendering DPI from Qt view
    for view in scene.views():
        RENDER_DPI['X'] = view.physicalDpiX()
        RENDER_DPI['Y'] = view.physicalDpiY()
        break

    # Go recursive first: render children
    count = 0
    for aname, agraph in graphtree['children'].items():
        LOG.debug("Rendering child scene: " + aname)
        # Render each child in a temporary scene to get the size of the scene
        # in order to resize the parent node accordingly
        temp_scene = type(scene)()
        render_statechart(temp_scene, agraph, keep_pos, dump_gfx)
        w = temp_scene.width()
        for node in graphtree['graph'].iternodes():
            if node == aname:
                size = temp_scene.itemsBoundingRect()
                node.attr['width'] = ((temp_scene.width() + 35)
                                      / RENDER_DPI['X'])
                node.attr['height'] = ((temp_scene.height() + 35)
                                      / RENDER_DPI['Y'])
                graphtree['children'][aname]['scene'] = temp_scene
                LOG.debug ("done")
                break

    # Harmonize the size of states to avoid having huge composite state(s)
    # next to single, small states. Rule: there can't be a state with a size
    # that is less than a third of the biggest state.
    try:
        min_width = float(max(float(node.attr.get('width', 0.0) or 0.0)
                        for node in graphtree['graph'].iternodes()))
        min_height = float(max(float(node.attr.get('height', 0.0) or 0.0)
                        for node in graphtree['graph'].iternodes()))
    except ValueError as err:
        LOG.debug(str(err))
        LOG.debug(traceback.format_exc())
        min_width, min_height = 0, 0
    if min_width and min_height:
        for node in graphtree['graph'].iternodes():
            if node.attr['shape'] != 'record':
                continue
            node.attr['width']  = node.attr.get('width') or (min_width / 3.0)
            node.attr['height'] = node.attr.get('height') or (min_height / 3.0)

    # Statechart symbols lookup table
    lookup = {'point'  : Point,
              'record' : Record,
              'diamond': Diamond,
              'square' : Stop}
    try:
        # Bonus: the tool can render any dot graph...
        graph = graphtree.get('graph', None) or dotgraph.AGraph('taste.dot')
        config = " ".join("{name}={val}".format(name=name, val=val)
                for name, val in graphtree['config'].items())
    except IOError as err:
        LOG.info('No statechart to display....')
        LOG.debug(str(err))
        raise
    graph.graph_attr.update(dpi='72.0')
    EDGES[:] = graph.edges()

    for each in scene.visible_symb:
        each.setVisible(False)
    #scene.clear()
    #G_SYMBOLS.clear()

    # Compute all the coordinates (self-modifying function)
    # Force the fontsize of the nodes to be 12, as in OpenGEODE
    # use -n1 below to keep user-specified node coordinates
    if dump_gfx:
        dump_name = 'sc_' + os.path.basename(dump_gfx)
        dump_gfx = os.path.dirname(dump_gfx) or '.' + os.sep + dump_name
        if dump_gfx.split('.')[-1].lower() != 'png':
            dump_gfx += '.png'

    graph.layout(prog='neato', args='{cfg} {kp} {dump}'
            .format(cfg=config, kp='-n1' if keep_pos else '',
                    dump=('-Tpng -o' + dump_gfx) if dump_gfx else ''))
    # bb is not visible directly - extract it from the low level api:
    bounding_rect = [float(val) for val in
            dotgraph.graphviz.agget(graph.handle, b'bb').split(b',')]
    dot_dpi = float(dotgraph.graphviz.agget(graph.handle, b'dpi'))

    #fontname = graph.graph_attr.get('fontname')
    #fontsize = graph.graph_attr.get('fontsize')
    #with open('statechart.dot', 'w') as output:
    #    output.write(graph.to_string())

    nodes = preprocess_nodes(graph, bounding_rect, dot_dpi)
    node_symbols = []
    for node in nodes:
        shape = node.get('shape')
        try:
            node_symbol = lookup[shape](node, graph)
            if str(node_symbol) in graphtree['children'] \
                    and shape == 'record':
                # Use a different color for non-terminal states
                node_symbol.setBrush(QBrush(QColor(249, 249, 249)))
            G_SYMBOLS.add(node_symbol)
            node_symbols.append(node_symbol)
            scene.addItem(node_symbol)
        except KeyError:
            raise TypeError('Statechart - unsupported shape: ' + shape)
    edges = preprocess_edges(graph, node_symbols, bounding_rect, dot_dpi)

    for edge in edges:
        Edge(edge, graph)

    # Make sure the scene has no negative coordinates
    scene.translate_to_origin()

    for aname, agraph in graphtree['children'].items():
        # At the end, place the content of the scene of the composite states
        # in the symbol by moving them from their temporary scene
        for symb in scene.visible_symb:
            if str(symb) == aname:
                deltapos = symb.scenePos() + QPointF(30.0, 30.0)
                for each in agraph['scene'].floating_symb:
                    # In principle we should change the parentItem to make sure
                    # that all children items are moved together with their
                    # parent. Unfortunately calls to setParentItem provoke
                    # a segfault. To be tried again when PySide is fixed...
                    # Update (04/2019) Workaround found (see sdlSymbols.py)
                    if symb.scene() != each.scene():
                        symb.scene().addItem(each)
                    each.setParent(symb)
                    each.position += deltapos
                    each.setZValue(each.zValue() + symb.zValue() + 1)


def lock():
    ''' Prevent multiple callers to render at the same time '''
    global g_statechart_lock
    g_statechart_lock = True

def unlock():
    ''' Prevent multiple callers to render at the same time '''
    global g_statechart_lock
    g_statechart_lock = False

def locked():
    ''' Return the lock status '''
    return g_statechart_lock


def create_dot_graph(root_ast,
                     basic=False,
                     scene=None,
                     view=None,
                     is_root=True,
                     config=None):
    ''' Return a dot.AGraph item, from an ogAST.Process or child entry
        Set basic=True to generate a simple graph with at most one edge
        between two states and no diamond nodes
        is_root is set to False for nested diagrams
    '''
    graph = dotgraph.AGraph(strict=False, directed=True)
    ret = {'graph': graph, 'children': {}, 'config': {}}
    diamond = 0

    # Define the list of input signals including timers
    if is_root:
        input_signals = {sig['name'].lower() for sig in root_ast.input_signals}
        for each in root_ast.timers:
            input_signals.add(each)
    else:
        # set by recursive caller:
        input_signals = root_ast.all_signals

    # Add the Connect parts below nested states, and continuous signals
    for each in root_ast.content.states:
        for connect in each.connects:
            input_signals |= set(connect.connect_list)
        for cs in each.continuous_signals:
            LOG.debug("Continuous signal: " + cs.trigger.inputString)
            input_signals.add("[" + cs.trigger.inputString.strip() + "]")

    # valid_inputs: list of messages to be displayed in the statecharts
    # user can remove them from the file to make cleaner diagrams
    # config_params can be set to tune the call to graphviz
    valid_inputs = set()
    config_params = {}
    identifier = getattr(root_ast, "statename", root_ast.processName)
    #LOG.info("Statechart: rendering scene " + identifier)
    #LOG.info("Input signals from model: " + str (input_signals))

    try:
        if not is_root:
            # Read config file only for the top-level diagram
            config_params = config
            # Adjust edge len for inner states. 1 is too small
            config_params["-Elen"] = "1.5"
            raise IOError
        with open (identifier + ".cfg", "r") as cfg_file:
            all_lines = (line.strip() for line in cfg_file.readlines())
        for each in all_lines:
            split = each.split()
            if len(split) == 3 and split[0] == "cfg":
                config_params[split[1]] = split[2]
            elif each:
                valid_inputs.add(each.lower())
    except IOError:
        valid_inputs = input_signals
        config_params = config_params or {"-Nfontsize" : "14",
                         "-Efontsize" : "8",
                         "-Gsplines"  : "curved",
                         "-Gsep"      : "0.3",
                         "-Gdpi"      : "72",
                         "-Gstart"    : "random10",
                         "-Goverlap"  : "scale",
                         "-Nstyle"    : "rounded",
                         "-Nshape"    : "record",
                         "-Elen"      : "1"}
        LOG.info ("... valid signals: " + ", ".join(valid_inputs))
    else:
        LOG.info ("Statechart settings read from " + identifier + ".cfg")
        LOG.info ("... using signals: " + ", ".join(valid_inputs))

    #LOG.debug(str(config_params))

    if scene and view:
        # Load and display a table for the user to filter out messages that
        # are not relevant to display on the statechart - and make it lighter
        # Do not repeat for substates (view is None and no cfg is saved)
        lock()
        def right(leftList, rightList):
            for each in leftList.selectedItems():
                item = leftList.takeItem(leftList.row(each))
                rightList.addItem(item)
        def left(leftList, rightList):
            for each in rightList.selectedItems():
                item = rightList.takeItem(rightList.row(each))
                leftList.addItem(item)
        loader = QUiLoader()
        ui_file = QFile(":/statechart_cfg.ui")
        ui_file.open(QFile.ReadOnly)
        dialog = loader.load(ui_file)
        dialog.setParent (view, Qt.Dialog)
        okButton = dialog.findChild(QPushButton, "okButton")
        rightButton = dialog.findChild(QToolButton, "toRight")
        leftButton = dialog.findChild(QToolButton, "toLeft")
        rightList = dialog.findChild(QListWidget, "rightList")
        leftList = dialog.findChild(QListWidget, "leftList")
        okButton.pressed.connect(dialog.accept)
        rightButton.pressed.connect(partial(right, leftList, rightList))
        leftButton.pressed.connect(partial(left, leftList, rightList))
        ui_file.close()
        rightList.addItems(list(valid_inputs))
        leftList.addItems(list(input_signals - valid_inputs))
        go = dialog.exec()
        valid_inputs.clear()
        for idx in range(rightList.count()):
            valid_inputs.add(rightList.item(idx).text())
        unlock()

    inputs_to_save = set(valid_inputs)
    for state in root_ast.mapping.keys():
        # create a new node for each state (including nested states)
        if state.endswith('START'):
            if len(state) > 5:
                label=state[0:-6]
            else:
                label=''
            graph.add_node(state,
                           label=label,
                           shape='point',
                           fixedsize='true',
                           width=10.0 / 72.0)
        else:
            # Find the path to the state, this is needed to be able to
            # highlight properly the parallel states in simulation
            for s in root_ast.content.states:
                # State may have commas or *
                statelist = [a.lower() for a in s.statelist]
                if state.lower() in statelist:
                    path = s.path
                    break
            else:
                # Should never happen
                LOG.error('Bug in statechart rendering - please report')
                path = []
            path_str = ".".join(elem.split()[1] for elem in path + [f"S {state}"]
                                if not elem.startswith("PROCESS"))
            graph.add_node(state,
                           label=state,
                           comment=path_str,
                           shape='record',
                           style='rounded')

    for each in [term for term in root_ast.terminators
                 if term.kind == 'return']:
        # create a new node for each RETURN statement (in nested states)
        ident = each.inputString.lower() or ' '
        graph.add_node(ident,
                       label='X', #ident, (no, otherwise size isn't ok)
                       shape='square',
                       width=1.0 / 72.0)

    # the AST does not contain a mapping the Connect parts below a state
    # Create it here on the spot
    connect_mapping = defaultdict(list)
    for each in root_ast.content.states:
        for stateName in each.statelist:
            connect_mapping[stateName.lower()].extend(each.connects)

    for state, inputs in chain(root_ast.mapping.items(),
                               root_ast.cs_mapping.items(),
                               connect_mapping.items()):
        # Add edges
        transitions = \
            inputs if not state.endswith('START') \
                   else [root_ast.transitions[inputs]]
        # Allow simplified graph, without diamonds and with at most one
        # transition from a given state to another
        target_states = defaultdict(set)
        for trans in transitions:
            source = state
            # transition label - there can be several inputs
            try:
                # Keep only message name, remove params and newlines
                # (newlines are not supported by graphviz)
                LOG.debug("Transition trigger: " + trans.inputString.strip())
                is_cs = "[" + trans.inputString.split(';')[0].strip() + "]"
                if is_cs in valid_inputs:
                    #  Make sure continuous signal strings are not filtered
                    #  split on semicolon because of the "priority" statement
                    label = is_cs #"[" + trans.inputString.strip() + "]"
                else:
                    label, = re.match(r'([^(]+)', trans.inputString).groups()
                    label = label.strip().replace('\n', ' ')
            except AttributeError as err:
                LOG.debug ("Start transition: " + str(err))
                # START transition may have no inputString
                for each in root_ast.content.named_start:
                    # each is of type ogAST.Start
                    if each.transition == trans:
                        label = each.inputString[:-6]
                        break
                else:
                    label = ''

            def find_terminators(trans):
                ''' Recursively find all NEXTSTATES and RETURN nodes '''
                next_states = [term for term in trans.terminators
                               if term.kind in ('next_state', 'return')]
                joins = [term for term in trans.terminators
                         if term.kind == 'join']
                for join in joins:
                    # JOIN - Find corresponding label
                    try:
                        corr_label, = [lab for lab in
                                       root_ast.content.floating_labels +
                                       root_ast.labels if
                                       lab.inputString.lower() ==
                                       join.inputString.lower()]
                    except ValueError:
                        LOG.error('Missing label: ' + join.inputString)
                    else:
                        # Don't recurse forever in case of livelock
                        try:
                            if corr_label.inputString != trans.inputString:
                                next_states.extend(find_terminators(corr_label))
                        except AttributeError:
                            # START transition -> no inputString
                            pass
                return set(next_states)

            # Determine the list of terminators in this transition
            next_states = find_terminators(trans)

            if len(next_states) > 1 and not basic:
                # more than one terminator - add intermediate node
                graph.add_node(str(diamond),
                               shape='diamond',
                               fixedsize='true',
                               width=15.0 / 72.0,
                               height=15.0 / 72.0, label='')
                if label.lower() in valid_inputs or not label.strip():
                    graph.add_edge(source,
                                   str(diamond),
                                   label=label)
                    inputs_to_save.add(label.lower())
                source = str(diamond)
                label = ''
                diamond += 1
            for term in next_states:
                if term.inputString.strip() in ('-', '-*'):
                    target = state
                else:
                    target = term.inputString.lower() or ' '
                if basic:
                    target_states[target] |= set(label.split(','))
                else:
                    labs = set(lab.strip() for lab in label.split(',') if
                                    lab.strip().lower() in valid_inputs | {""})
                    actual = ',\n'.join(labs)
                    graph.add_edge(source,
                                   target,
                                   label=actual)
                    inputs_to_save |= set(lab.lower() for lab in labs)
        for target, labels in target_states.items():
            # Here we add the label if in valid_inputs.
            sublab = [lab.strip() for lab in labels if
                      lab.strip().lower() in valid_inputs | {""}]
            # Basic mode
            if sublab:
                label=',\n'.join(sublab)
                inputs_to_save |= set(lab.lower() for lab in sublab)
            else:
                label = ''
            graph.add_edge(source,
                           target,
                           label=label)
    # Uncomment for debugging the generated dot graph:
    #with open('statechart.dot', 'w') as output:
    #   output.write(graph.to_string())
    #return graph
    if is_root:
        with open(identifier + ".cfg", "w") as cfg_file:
            for name, value in config_params.items():
                cfg_file.write("cfg {} {}\n".format(name, value))
            for each in inputs_to_save:
                cfg_file.write(each + "\n")
    ret['config'] = config_params
    for each in root_ast.composite_states:
        LOG.debug ("Recursive generation of statechart: " + each.statename)
        # Recursively generate the graphs for nested states
        # Inherit from the list of signals from the higher level state
        #each.input_signals = root_ast.input_signals
        each.all_signals = valid_inputs
        ret['children'][each.statename] = create_dot_graph(each,
                                                          basic,
                                                          scene,
                                                          is_root=False,
                                                          config=config_params)
    return ret


if __name__ == '__main__':
    render_statechart(None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    OpenGEODE - A tiny, free SDL Editor for TASTE

    SDL is the Specification and Description Language (Z100 standard from ITU)

    Copyright (c) 2012-2022 European Space Agency

    Designed and implemented by Maxime Perrotin

    Contact: maxime.perrotin@esa.int

    This module is responsible for transforming AST elements to actual symbols

    It is separated from the main SDL_Scene class as the rendering can
    be done on any scene (e.g. clipboard).

    There is a single rendering function for all SDL construct, and a dispatch
    mechanism (using the Python3-backported feature called singledispatch).

    Rendering can be done for single elements (returns the symbol) or for
    complete diagrams.

    This rendering capability is separated from the AST definition (ogAST.py)
    so that the AST module is kept independent from any graphical backend and
    is not related to Pyside.

    When rendering a (set of) symbol(s), update text autocompletion list(s).
"""

import logging
from itertools import chain
from functools import singledispatch

from .ogParser import type_name
from . import ogAST, sdlSymbols, genericSymbols


LOG = logging.getLogger(__name__)

__all__ = ['render', 'add_to_scene']


def add_to_scene(item, scene):
    ''' Add item to a scene after verifying that the scene allows it '''
    if type(item) in scene.allowed_symbols:
        scene.addItem(item)
    else:
        #print type(item), scene.allowed_symbols, scene.context
        raise TypeError('This symbol does not fit the current scene')


@singledispatch
def render(ast, scene, parent, states, terminators=None):
    ''' Render a transition action symbol on the scene '''
    _, _, _, _ = scene, parent, states, terminators
    # Default behaviour is to raise an exception, if there is no
    # rendering function for a given symbol. Otherwise the dispatch
    # mechanism forwards the call to a registered function (see below)
    raise TypeError('[Renderer] Unsupported symbol in branch: ' + repr(ast))


@render.register(ogAST.Block)
def _block(ast, scene):
    ''' Render a block, containing a set of process symbols '''
    top_level = []
    for each in ast.processes:
        top_level.append(render(each, scene))
        if each.instance_of_ref:
            top_level.append(render(each.instance_of_ref, scene))
    for each in ast.parent.text_areas:
        # System level may contain text areas with signal definitions, etc.
        top_level.append(render(each, scene))
    if not ast.parent.text_areas:
        # If signals are declared outside from a textbox, create one
        signals = ["signal {si[name]}{param};\n".format(si=sig,
           param=('(' + sig['type'].ReferencedTypeName.replace('-', '_') + ')')
                 if 'type' in sig else '')
             for sig in ast.parent.signals]
        procedures = ["{exported}procedure {proc.inputString};\n{optfpar}{external};\n"
                      .format(proc=proc,
                              exported="exported " if proc.exported else "",
                              external="referenced" if proc.referenced else "external",
                              optfpar="fpar\n    " + ",\n    ".join
                              (["{direc} {fp[name]} {asn1}"
                                .format(fp=fpar,
                                        direc="in"
                                           if fpar['direction']=='in'
                                           else 'in/out',
                                        asn1=getattr(fpar['type'],
                                           'ReferencedTypeName', 'TYPE_ERROR')
                                           .replace('-', '_'))
                                for fpar in proc.fpar]) + ';\n'
                                    if proc.fpar else '')
                        for proc in ast.parent.procedures]
        if signals or procedures:
            text_area = ogAST.TextArea()
            text_area.inputString = "{}\n\n{}".format('\n'.join(signals),
                                                      '\n'.join(procedures))
            text_area.pos_x = scene.itemsBoundingRect().width()
            text_area.pos_y = scene.itemsBoundingRect().y() + 10
            top_level.append(render(text_area, scene))
    return top_level


@render.register(ogAST.Process)
def _process(ast, scene, **_):
    ''' Render a Process symbol (in a BLOCK diagram) '''
    if ast.process_type:
        symbol = sdlSymbols.ProcessType(ast, ast)
    else:
        symbol = sdlSymbols.Process(ast, ast)
    add_to_scene(symbol, scene)
    return symbol


@render.register(ogAST.Automaton)
def _automaton(ast, scene):
    ''' Render graphical elements of a process or procedure '''
    top_level_symbols = []
    # Render text areas (DCL declarations, etc.)
    for text in ast.textAreas:
        top_level_symbols.append(render(text, scene))

    # Render procedures symbols
    top_level_symbols.extend(
            [render(proc, scene)
                for proc in ast.inner_procedures
               if not proc.external and not proc.textual_procedure])

    # Render the start symbol
    if ast.start:
        top_level_symbols.append(render(ast.start, scene, ast.states))

    # Render named start symbols in nested states
    for each in ast.named_start:
        top_level_symbols.append(render(each, scene, ast.states))

    # Render floating labels
    for label in ast.floating_labels:
        top_level_symbols.append(render(label, scene, ast.states))

    # Render floating states
    nested_states = []
    for state in ast.states:
        # Create only floating states
        try:
            new_state = render(state, scene=scene, states=ast.states,
                               terminators=ast.parent.terminators)
            if new_state.nested_scene:
                if str(new_state).lower() in nested_states:
                    new_state.nested_scene = None
                else:
                    nested_states.append(str(new_state).lower())
        except TypeError:
            # Discard terminators (see _state function for explanation)
            pass
        else:
            top_level_symbols.append(new_state)

    # If the source .pr contained FPAR outside a textbox, create one
    if ast.parent.fpar and not any(x.fpar for x in ast.textAreas):
        text_area = ogAST.TextArea()
        fpars = ('{} {}'.format(fp['name'],
                                type_name(fp['type']).replace('-', '_'))
                    for fp in ast.parent.fpar)
        text_area.inputString = ("-- Formal parameters\n"
                                "fpar {};".format(',\n          '.join(fpars)))
        text_area.pos_x = scene.itemsBoundingRect().x() - 200
        text_area.pos_y = scene.itemsBoundingRect().y()
        top_level_symbols.append(render(text_area, scene))
    return top_level_symbols


@render.register(ogAST.State)
def _state(ast, scene, states, terminators, parent=None):
    ''' Render a floating state and its inputs '''
    _ = parent
    # Discard the state if it is a terminator too as it is not a floating
    # state in that case: it will be rendered together with all its (possible)
    # INPUT children in the render_terminator function.
    for term in terminators:
        state_label = ast.via or ast.inputString
        if(term.kind == 'next_state' and
                term.pos_x == ast.pos_x and
                term.pos_y == ast.pos_y and
                term.inputString == state_label):
            raise TypeError('This state is a terminator')
    new_state = sdlSymbols.State(parent=None, ast=ast)
    if new_state not in scene.items():
        add_to_scene(new_state, scene)

    for exit in chain(ast.inputs, ast.connects, ast.continuous_signals):
        render(exit, scene=scene, parent=new_state, states=states)

    new_state.nested_scene = ast.composite or ogAST.CompositeState()

    return new_state


@render.register(ogAST.Procedure)
def _procedure(ast, scene, parent=None, states=None):
    ''' Add a procedure symbol to the scene '''
    _, _ = parent, states
    proc_symbol = sdlSymbols.Procedure(ast, ast)
    add_to_scene(proc_symbol, scene)
    return proc_symbol


@render.register(ogAST.TextArea)
def _text_area(ast, scene, parent=None, states=None):
    ''' Render a text area from the AST '''
    _, _ = parent, states
    text = sdlSymbols.TextSymbol(ast)
    add_to_scene(text, scene)
    return text


@render.register(ogAST.Start)
def _start(ast, scene, states, parent=None):
    ''' Add the start symbol to a scene '''
    _ = parent
    start_symbol = sdlSymbols.Start(ast)
    add_to_scene(start_symbol, scene)
    if ast.transition:
        render(ast.transition, scene=scene, parent=start_symbol, states=states)
    return start_symbol


@render.register(ogAST.CompositeState_start)
def _composite_start(ast, scene, states, parent=None):
    ''' Add an editable start symbol to a scene (in composite states) '''
    _ = parent
    start_symbol = sdlSymbols.StateStart(ast)
    add_to_scene(start_symbol, scene)
    if ast.transition:
        render(ast.transition,
                      scene=scene, parent=start_symbol, states=states)
    return start_symbol


@render.register(ogAST.Procedure_start)
def _procedure_start(ast, scene, states, parent=None):
    ''' Add the procedure start symbol to a scene '''
    _ = parent
    start_symbol = sdlSymbols.ProcedureStart(ast)
    add_to_scene(start_symbol, scene)
    if ast.transition:
        render(ast.transition, scene=scene, parent=start_symbol, states=states)
    return start_symbol


@render.register(ogAST.Floating_label)
def _floating_label(ast, scene, states, parent=None):
    ''' Add a Floating label from the AST to the scene '''
    _ = parent
    lab = sdlSymbols.Label(parent=None, ast=ast)
    if lab not in scene.items():
        add_to_scene(lab, scene)
    lab.pos_x, lab.pos_y = ast.pos_x, ast.pos_y
    if ast.transition:
        render(ast.transition, scene=scene, parent=lab, states=states)
    return lab


@render.register(ogAST.Transition)
def _transition(ast, scene, parent, states):
    ''' Add a transition to a scene '''
    for each in ast.actions:
        # pylint: disable=E1111
        parent = render(each, scene=scene, parent=parent, states=states)

    if ast.terminator:
        render(ast.terminator, scene=scene, parent=parent, states=states)


@render.register(ogAST.Comment)
def _comment(ast, scene, parent, states=None):
    ''' Create a COMMENT symbol - note: relative positioning is lost '''
    _, _ = scene, states
    return genericSymbols.Comment(parent, ast=ast)


@render.register(ogAST.Task)
def _task(ast, scene, parent, states):
    ''' Create a TASK symbol '''
    _, _ = scene, states
    return sdlSymbols.Task(parent, ast=ast)


@render.register(ogAST.Output)
def _output(ast, scene, parent, states):
    ''' Create an OUTPUT or PROCEDURE CALL symbol '''
    _, _ = scene, states
    return sdlSymbols.Output(parent, ast=ast)


@render.register(ogAST.ProcedureCall)
def _procedure_call(ast, scene, parent, states):
    ''' Create an OUTPUT or PROCEDURE CALL symbol '''
    _, _ = scene, states
    return sdlSymbols.ProcedureCall(parent, ast=ast)


@render.register(ogAST.Decision)
def _decision(ast, scene, parent, states):
    ''' Create a DECISION symbol and all its answers '''
    symbol = sdlSymbols.Decision(parent, ast=ast)
    # Place the symbol at absolute coordinates
    if not parent:
        symbol.pos_x, symbol.pos_y = ast.pos_x, ast.pos_y
    for branch in ast.answers:
        render(branch, scene=scene, parent=symbol, states=states)
    symbol.updateConnectionPointPosition()
    return symbol


@render.register(ogAST.Label)
def _label(ast, scene, parent=None, states=None):
    ''' Create a LABEL symbol '''
    _, _ = scene, states
    return sdlSymbols.Label(parent, ast=ast)


@render.register(ogAST.Answer)
def _answer(ast, scene, parent, states):
    ''' Create an ANSWER symbol and build its following transition '''
    symbol = sdlSymbols.DecisionAnswer(parent, ast=ast)
    # Place the symbol at absolute coordinates so that if
    # the branch has NEXTSTATEs symbols, they are properly placed
    if not parent:
        symbol.pos_x, symbol.pos_y = ast.pos_x, ast.pos_y
    if ast.transition:
        render(ast.transition, scene=scene, parent=symbol, states=states)
    return symbol


@render.register(ogAST.Terminator)
def _terminator(ast, scene, parent, states):
    ''' Create a TERMINATOR symbol '''
    if ast.label:
        # pylint: disable=E1111
        parent = render(ast.label, scene=scene, parent=parent, states=states)
    if ast.kind == 'next_state':
        # Create a new state symbol
        symbol = sdlSymbols.State(parent=parent, ast=ast)
        # If the terminator is also a new state, render the inputs below
        for state_ast in states:
            state_label = state_ast.via or state_ast.inputString
            if (state_label == ast.inputString and
                    state_ast.pos_x == ast.pos_x and
                    state_ast.pos_y == ast.pos_y):
                symbol.nested_scene = state_ast.composite or \
                                      ogAST.CompositeState()
                for each in chain(state_ast.inputs,
                                  state_ast.connects,
                                  state_ast.continuous_signals):
                    render(each, scene=scene, parent=symbol, states=states)
                break
        symbol.nested_scene = ast.composite or ogAST.CompositeState()
    elif ast.kind == 'join':
        symbol = sdlSymbols.Join(parent, ast)
    elif ast.kind in ('return', 'stop'):
        symbol = sdlSymbols.ProcedureStop(parent, ast)
    else:
        raise TypeError('Unsupported terminator: ' + repr(ast))
    return symbol


@render.register(ogAST.Input)
def _input(ast, scene, parent, states):
    ''' Add input from the AST to the scene '''
    # Note: PROVIDED clause is not supported
    inp = sdlSymbols.Input(parent, ast=ast)
    if inp not in scene.items():
        add_to_scene(inp, scene)
    if not parent:
        inp.pos_x, inp.pos_y = ast.pos_x, ast.pos_y
    if ast.transition:
        render(ast.transition,
               scene=scene,
               parent=inp,
               states=states)
    return inp


@render.register(ogAST.ContinuousSignal)
def _continuous_signal(ast, scene, parent, states):
    ''' Add continuous signal to the scene '''
    if ast.artificial:
        return None
    cont = sdlSymbols.ContinuousSignal(parent, ast=ast)
    if cont not in scene.items():
        add_to_scene(cont, scene)
    if not parent:
        cont.pos_x, cont.pos_y = ast.pos_x, ast.pos_y
    if ast.transition:
        render(ast.transition,
               scene=scene,
               parent=cont,
               states=states)
    return cont


@render.register(ogAST.Connect)
def _connect(ast, scene, parent, states):
    ''' Add connect symbol from the AST to the scene '''
    conn = sdlSymbols.Connect(parent, ast=ast)
    if conn not in scene.items():
        add_to_scene(conn, scene)
    if not parent:
        conn.pos_x, conn.pos_y = ast.pos_x, ast.pos_y
    if ast.transition:
        render(ast.transition,
               scene=scene,
               parent=conn,
               states=states)
    return conn



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    OpenGEODE - A tiny, free SDL Editor for TASTE

    SDL is the Specification and Description Language (Z100 standard from ITU)

    Copyright (c) 2012-2019 European Space Agency

    Designed and implemented by Maxime Perrotin

    Contact: maxime.perrotin@esa.int

    This module is managing the Copy and Paste functions.
"""

import os
import traceback
import logging
from itertools import chain
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from . import ogAST, ogParser, sdlSymbols, genericSymbols, Renderer, Pr

__all__ = ['copy', 'paste']

LOG = logging.getLogger(__name__)

# Buffer holding the list of copied symbols in AST form
COPY_PASTE = []

# Actual scene clipboard
CLIPBOARD = None

# System clipboard used to copy-paste between instances of Opengeode
# Value is set when Opengeode is initialized
SYS_CLIPBOARD = None

def copy(selection):
    ''' Create a copy (duplicate) of the selected symbols in AST form '''
    # Clear the copy paste buffer
    COPY_PASTE[:] = []
    branch_top_level = []
    floating_items = []
    for item in selection:
        # When several items are selected, take the first of each subbranch
        if item.hasParent and not item.parentItem().grabber.isSelected():
            branch_top_level.append(item)
        elif not item.hasParent:
            # Take also floating items
            floating_items.append(item)
    # Check if selected items would allow a paste - reject copy otherwise
    # e.g. floating and non-floating items cannot be pasted together
    if ((branch_top_level == []) == (floating_items == []) or
            len(branch_top_level) > 1):
        raise TypeError('Selection is incompatible with copy')
    # Then parse/copy the selected branches
    for item in branch_top_level + floating_items:
        branch_ast, terminators = copy_branch(item)
        COPY_PASTE.append((branch_ast, terminators))
    LOG.debug('COPY-PASTE LIST: %s', str(COPY_PASTE))


def copy_branch(top_level_item):
    ''' Copy branches (recursively) '''
    res_terminators = []
    # Parse without symbol ids, so that pos_x is kept with the right value
    Pr.g_useSymbolId = False
    pr_text = '\n'.join(Pr.generate(top_level_item, cpy=True,
                                    nextstate=False, recursive=True))
    item_ast, terminators = top_level_item.get_ast(pr_text)
    LOG.debug('COPY ' + str(item_ast))
    if not item_ast:
        LOG.error('ERROR - copy failed')
        return

    branch = [item_ast]

    # If top_level_item is an horizontal symbol, all its children (branch)
    # are automatically copied. In the case that several following symbols
    # are selected however, the followers of the top_level must be
    # explicitly added to the copy list:
    if not isinstance(top_level_item, genericSymbols.HorizontalSymbol):
        next_aligned = top_level_item.next_aligned_symbol()
        while next_aligned and next_aligned.grabber.isSelected():
            next_pr_text = '\n'.join(Pr.generate(next_aligned, cpy=True,
                                            nextstate=False, recursive=True))
            next_ast, next_terminators = next_aligned.get_ast(next_pr_text)
            terminators.extend(next_terminators)
            branch.append(next_ast)
            next_aligned = next_aligned.next_aligned_symbol()

    # Parse terminators recursively (e.g. NEXTSTATE)
    res_terminators = terminators
    for term in terminators:
        # Get symbol at terminator coordinates
        symbols = top_level_item.scene().items(QRectF
                (term.pos_x, term.pos_y, term.width, term.height).center())
        for symbol in symbols:
            if (isinstance(symbol, sdlSymbols.State) and [c for c in
                symbol.childSymbols() if isinstance(c, (sdlSymbols.Input,
                    sdlSymbols.Connect, sdlSymbols.ContinuousSignal))]):
                term_branch, term_inators = copy_branch(symbol)
                branch.extend(term_branch)
                res_terminators.extend(term_inators)

    if SYS_CLIPBOARD is not None:
        #  Basic copy of a single branch to the system clipboard
        ident = top_level_item.common_name \
                if not isinstance(top_level_item, sdlSymbols.State) \
                else 'state'
        SYS_CLIPBOARD.setText("OG_SDL@-@{}@-@{}@-@{}".format(os.getpid(),
                                                         ident,
                                                         pr_text))
    return branch, res_terminators


def paste(parent, scene):
    '''
        Paste previously copied symbols at selection point
    '''
    remove_after_paste = False
    if SYS_CLIPBOARD is not None:
        #  Get from clipboard only if it comes from another process
        #  otherwise use the local clipboard which contains more data
        #  (for the moment the shared clipboard only stores top-level items)
        shared = SYS_CLIPBOARD.text().split('@-@') or []
        if len(shared) >= 4 \
                and shared[0] == "OG_SDL" \
                and shared[1] != str(os.getpid()):
            common_name = shared[2]
            pr_text     = shared[3]
            #  Copy to the local clipboard
            # LOG.debug("PASTE: " + pr_text)
            ast, _, _, _, terminators = \
                ogParser.parseSingleElement(common_name, pr_text)
            COPY_PASTE.append(([ast], terminators))
            remove_after_paste = True

    CLIPBOARD.clear()
    if not parent:
        new_symbols = paste_floating_objects(scene)
        for each in new_symbols:
            #  Make sure nested scenes are set properly
            each.double_click()
    else:
        new_symbols = paste_below_item(parent, scene)
        def rec_double_click(next):
            if next is not None:
                next.double_click()
                rec_double_click(next.next_aligned_symbol())
        for each in new_symbols:
            #  Make sure nested scenes are set properly
            rec_double_click(each)

    if remove_after_paste:
        #  Remove from local clipboard if it came from system clipboard
        #  otherwise it would be added over and over
        COPY_PASTE.pop()
    return new_symbols


def paste_floating_objects(scene):
    ''' Paste items with no parents (states, text areas) '''
    symbols = []
    LOG.debug('PASTING FLOATING OBJECTS')

    for item_list, terminators in COPY_PASTE:
        # Check if there are NEXTSTATE terminators while we try to paste
        # inside a prodcedure (this is forbidden, procedures have no states)
        for term in terminators:
            if term.kind == 'next_state' and scene.context == 'procedure':
                raise TypeError("Cannot copy states inside a procedure")

        # states is a list passed as parameter - not a generator:
        start = [i for i in item_list if isinstance(i, ogAST.Start)]
        states = [i for i in item_list if isinstance(i, ogAST.State)]
        text_areas = (i for i in item_list if isinstance(i, ogAST.TextArea))
        labels = (i for i in item_list if isinstance(i, ogAST.Floating_label))
        procedures = (i for i in item_list if isinstance(i, ogAST.Procedure))
        processes = (i for i in item_list if isinstance(i, ogAST.Process))
        for state in states:
            # nesting is not copied, because the nested content is a scene
            # that is just referenced.. content must be actually copied
            # to a new scene only if a state is renamed (this will preserve
            # the nesting content of the original state)

            # First check if state has already been pasted
            try:
                new_item = Renderer.render(state, scene=CLIPBOARD,
                           terminators=terminators, states=states)
            except TypeError as err:
                LOG.debug('No paste "' + state.inputString + '" -' + str(err))
                # Discard terminators (explanation given in Renderer._state)
                pass
            else:
                LOG.debug('PASTE STATE "' + state.inputString + '"')
                symbols.append(new_item)
                # Insert the new state at click coordinates
                Renderer.add_to_scene(new_item, scene)
        for each in chain(text_areas, labels):
            LOG.debug('PASTE Text Area/Label')
            new_item = Renderer.render(each, scene, states=states)
            symbols.append(new_item)
        for each in chain(procedures, processes):
            LOG.debug('PASTE Process/Procedure')
            new_item = Renderer.render(each, scene, states=states)
            symbols.append(new_item)
            new_item.nested_scene = scene.create_subscene(
                                                        new_item.context_name)
            # Render recursively, creating any required scene
            try:
                new_item.nested_scene.render_everything(each.content)
            except TypeError as err:
                LOG.debug(str(err))

        if start:
            start, = start
            LOG.debug('PASTE START')
            for item in scene.visible_symb:
                if isinstance(item, sdlSymbols.Start):
                    raise TypeError('Only one START symbol is possible')
            new_item = Renderer.render(start, scene, states=states)
            symbols.append(new_item)
    return symbols


def paste_below_item(parent, scene):
    ''' Paste items under a selected symbol '''
    LOG.debug('Pasting below item ' + repr(parent)[slice(0, 20)])
    symbols = []
    for item_list, terminators in COPY_PASTE:

        # Check if there are NEXTSTATE terminators while we try to paste
        # inside a prodcedure (this is forbidden, procedures have no states)
        for term in terminators:
            if term.kind == 'next_state' and scene.context == 'procedure':
                raise TypeError("Cannot copy states inside a procedure")

        states = [i for i in item_list if isinstance(i, ogAST.State)]
        for i in [c for c in item_list if not isinstance
                (c, (ogAST.State, ogAST.TextArea, ogAST.Start))]:
            LOG.debug('PASTE ' + str(i))
            # Create the new item from the AST description
            new_item = Renderer.render(i, scene=CLIPBOARD,
                                       parent=None, states=states)
            if isinstance(new_item, sdlSymbols.Label) \
                    and parent.next_aligned_symbol() is not None:
                # When a transition starting from a Label is pasted, we must
                # check that the last symbol is not a terminator unless the
                # parent had no following symbol. This would clash otherwise.
                follower = new_item.next_aligned_symbol()
                while follower is not None:
                    if isinstance(follower, (sdlSymbols.Join, sdlSymbols.State)):
                        raise TypeError("Cannot paste here because of the terminator symbol")
                    follower = follower.next_aligned_symbol()

            # Check that item is compatible with parent
            if (type(new_item).__name__ in parent.allowed_followers):
                # Move the item from the clipboard to the scene
                Renderer.add_to_scene(new_item, scene)
                new_item.pos_x = new_item.pos_y = 0.0
                symbols.append(new_item)
            else:
                raise TypeError('Cannot paste here ({t1} cannot follow {t2})'
                                .format(t1=type(new_item), t2=type(parent)))
    return symbols



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Diagram elements for connectors. Different types of connectors are
    available by default, such as lines, arrows, and bezier curves.

    The "Connection" class is the mother class, all other connectors must
    inherit from it and possibly redefine some functions or shape.

    Copyright (c) 2012-2020 European Space Agency

    Designed and implemented by Maxime Perrotin

    Contact: maxime.perrotin@esa.int
"""

import math
import logging

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *
from . import ogParser
from .TextInteraction import EditableText

LOG = logging.getLogger(__name__)


# pylint: disable=R0904
class Connection(QGraphicsPathItem):
    ''' Connection between two symbols (top-level class) '''
    # Default connections are not selectable
    default_cursor = Qt.ArrowCursor
    hasParent = False

    def __init__(self, parent, child):
        ''' Create a new connection between a parent and a child item '''
        super().__init__(parent)
        self.parent = parent
        self.child = child
        self._start_point = None
        self._end_point = None
        self._middle_points = []
        self.default_pen = QPen()
        self.default_pen.setColor(Qt.blue)
        self.default_pen.setCosmetic(False)
        self.setPen(self.default_pen)
        self.bold_pen = QPen()
        self.bold_pen.setWidth(3)
        self.parent_rect = parent.sceneBoundingRect()
        self.childRect = child.sceneBoundingRect()
        # Activate cache mode to boost rendering by calling paint less often
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
        # When the parent or child move, the connection may need
        # to adjust the end point: done upon signal reception
        self.child.moved.connect(self.child_moved)
        self.parent.moved.connect(self.parent_moved)
        # Syntax error indicator
        self.syntax_error: Boolean = False
        # Flag to choose the bold pen when painting the edge if selected
        self.selected = False

    @Slot(float, float)
    def child_moved(self, delta_x, delta_y):
        ''' When the connection child moves - redefine in subclasses '''
        pass

    @Slot(float, float)
    def parent_moved(self, delta_x, delta_y):
        ''' When the connection parent moves - redefine in subclasses '''
        self.parent.update_connections()

    @property
    def start_point(self):
        ''' Compute connection origin - redefine in subclasses '''
        return self._start_point or QPointF(0, 0)

    @property
    def end_point(self):
        ''' Compute connection end point - redefine in subclasses '''
        return self._end_point or QPointF(0, 0)

    @property
    def middle_points(self):
        ''' Compute connection intermediate points - redefine in subclasses '''
        return self._middle_points

    def top_level(self):
        ''' In case the connection has a text, implement top_level()
            as needed by TextInteraction '''
        return self.parent

    def simple_arrow(self, origin='head', path=None):
        ''' Compute the two points of an vertical arrow head '''
        if origin == 'head':
            endp = self.end_point
        else:
            endp = self.start_point
        return (QPointF(endp.x() - 5, endp.y() - 5),
                QPointF(endp.x() + 5, endp.y() - 5))

    def select(self, selected=True):
        ''' Select a connection - make it bold '''
        if selected:
            self.selected = True
            self.setPen(self.bold_pen)
        else:
            self.selected = False
            self.setPen(self.default_pen)
        self.update()   # force a repaint

    def angle_arrow(self, path, origin='head'):
        ''' Compute the two points of the arrow head with the right angle '''
        if origin == 'tail':
            path = path.toReversed()
        length = path.length()
        percent = path.percentAtLength(length - 10.0)
        src = path.pointAtPercent(percent)
        #path.moveTo(path.pointAtPercent(1))
        end_point = path.pointAtPercent(1)
        #end_point = path.currentPosition()
        line = QLineF(src, end_point)
        angle = math.acos(line.dx() / (line.length() or 1))
        if line.dy() >= 0:
            angle = math.pi * 2 - angle
        arrow_size = 10.0
        arrow_p1 = end_point + QPointF(
                math.sin(angle - math.pi / 3) * arrow_size,
                math.cos(angle - math.pi / 3) * arrow_size)
        arrow_p2 = end_point + QPointF(
                math.sin(angle - math.pi + math.pi / 3) * arrow_size,
                math.cos(angle - math.pi + math.pi / 3) * arrow_size)
        return (arrow_p1, arrow_p2)

    def draw_arrow_head(self, shape, origin='head', kind='simple'):
        ''' Generic function to draw a simple arrow '''
        if kind == 'simple':
            arrowhead = self.simple_arrow(path=shape)
        else:
            arrowhead = self.angle_arrow(shape, origin)
        shape.lineTo(arrowhead[0])
        shape.moveTo(self.end_point if origin == 'head' else self.start_point)
        shape.lineTo(arrowhead[1])

    def __str__(self):
        ''' Print connection information for debug purpose'''
        return 'Connection: parent = {p}, child = {c}'.format(
                p=str(self.parentItem()), c=str(self.child))

    def reshape(self):
        ''' Update the connection or arrow shape '''
        shape = QPainterPath()
        shape.moveTo(self.start_point)
        for point in self.middle_points:
            shape.lineTo(point)
        shape.lineTo(self.end_point)
        # If required draw an arrow head (e.g. in SDL NEXTSTATE and JOIN)
        if self.child.arrow_head:
            self.draw_arrow_head(shape, origin='head',
                                 kind=self.child.arrow_head)
        if self.child.arrow_tail:
            shape.moveTo(shape.pointAtPercent(0))
            self.draw_arrow_head(shape, origin='tail',
                                 kind=self.child.arrow_head)
        self.setPath(shape)


class RakeConnection(Connection):
    ''' Fork-like connection, e.g. between a state and an input symbol '''
    @property
    def start_point(self):
        ''' Compute connection origin - redefined function '''
        parent_rect = self.parentItem().boundingRect()
        return QPointF(parent_rect.width() / 2, parent_rect.height())

    @property
    def end_point(self):
        ''' Compute connection end point - redefined function '''
        coord = self.child.pos()
        coord.setX(coord.x() + self.child.boundingRect().width() / 2)
        return coord

    @property
    def middle_points(self):
        ''' Compute connection intermediate points - redefined function '''
        yield QPointF(self.start_point.x(), self.start_point.y() + 10)
        yield QPointF(self.end_point.x(), self.start_point.y() + 10)


class JoinConnection(Connection):
    ''' Inverted fork-like connection, to join to a common point '''
    def reshape(self):
        ''' Update the connection shape - redefined: if the last element
        of a branch is e.g. a nextstate, don't join the connection point '''
        if self.parentItem().terminal_symbol:
            self.setPath(QPainterPath())
        else:
            super().reshape()

    @property
    def start_point(self):
        ''' Compute connection origin - redefined function '''
        parent_rect = self.parentItem().boundingRect()
        return getattr(self.parentItem(), 'connectionPoint',
                       QPointF(parent_rect.width() / 2, parent_rect.height()))

    @property
    def end_point(self):
        ''' Compute connection end point - redefined function '''
        point_scene = self.child.mapToScene(self.child.connectionPoint)
        return self.mapFromScene(point_scene)

    @property
    def middle_points(self):
        ''' Compute connection intermediate points - redefined function '''
        yield QPointF(self.start_point.x(), self.end_point.y() - 10)
        yield QPointF(self.end_point.x(), self.end_point.y() - 10)


class VerticalConnection(Connection):
    ''' Vertical line with or without arrow '''
    @property
    def start_point(self):
        ''' Compute connection origin - redefined function '''
        parent_rect = self.parentItem().boundingRect()
        return getattr(self.parentItem(), 'connectionPoint',
                       QPointF(parent_rect.width() / 2, parent_rect.height()))

    @property
    def end_point(self):
        ''' Compute connection end point - redefined function '''
        point = self.child.pos()
        point.setX(self.start_point.x())
        return point


class CommentConnection(Connection):
    ''' Class handling connection to comment symbols, fixed at the right
        of a symbol '''
    @property
    def start_point(self):
        ''' Compute connection origin - redefined function '''
        parent_rect = self.parentItem().boundingRect()
        return QPointF(parent_rect.width(), parent_rect.height() / 2)

    @property
    def end_point(self):
        ''' Compute connection end point - redefined function '''
        if self.child.on_the_right:
            return QPointF(self.child.x(),
                   self.child.y() + self.child.boundingRect().height() / 2)
        else:
            return QPointF(self.child.x() + self.child.boundingRect().width(),
                   self.child.y() + self.child.boundingRect().height() / 2)

    @property
    def middle_points(self):
        ''' Compute connection intermediate points - redefined function '''
        # Make sure the connection does not overlap the comment item
        if (self.child.on_the_right or
               (not self.child.on_the_right and
               self.child.x() + self.child.boundingRect().width()
               < self.parentItem().boundingRect().width())):
            go_to_point = self.start_point.x() + 5
        else:
            go_to_point = self.end_point.x() + 5
        yield QPointF(go_to_point, self.start_point.y())
        yield QPointF(go_to_point, self.end_point.y())


class SignalList(EditableText):
    ''' Simplified text editor for signal lists '''
    def __init__(self, parent, text='', hyperlink = None):
        ''' Smaller font than normal text '''
        super().__init__(parent, text)
        self.setFont(QFont('Ubuntu', pointSize=8))

    def set_text_alignment(self):
        ''' Text justification - ignore '''
        pass

    def set_textbox_position(self):
        ''' Alignment vs parent - ignore, done by the parent '''
        pass

    def try_resize(self):
        ''' Resizing of the parent when size expands - ignore '''
        pass

    # pylint: disable=C0103
    def focusOutEvent(self, event):
        ''' Redefined function - update in_sig and out_sig in parent '''
        super().focusOutEvent(event)


class Signalroute(Connection):
    ''' Subclass of Connection used to draw channels between processes '''
    in_sig = out_sig = None
    completion_list = set()

    def __init__(self, parent, child=None):
        ''' Set generic parameters from Connection class '''
        super().__init__(parent, child or parent)
        self.parser = ogParser
        self.blackbold = ()
        self.redbold = ()
        self.label_in = SignalList(parent=self)
        self.label_out = SignalList(parent=self)
        if not Signalroute.in_sig:
            # keep at class level as long as only one process is supported
            # when copy-pasting a process the channel in/out signal lists
            # are not parsed. Workaround is to keep the list "global"
            # to allow a copy of both process and channel
            # Needed for the image exporter, that copies the scene to a
            # temporary one
            Signalroute.in_sig = '{}'.format(',\n'.join(sig['name']
                                   for sig in parent.input_signals))
            Signalroute.out_sig = '{}'.format(',\n'.join(sig['name']
                                    for sig in parent.output_signals))
        self.label_in.setPlainText('[{}]'.format(self.in_sig))
        self.label_out.setPlainText('[{}]'.format(self.out_sig))
        self.label_in.document().contentsChanged.connect(self.change_siglist)
        self.label_out.document().contentsChanged.connect(self.change_siglist)
        for each in (self.label_in, self.label_out):
            each.show()
        #self.reshape()

    @Slot()
    def change_siglist(self):
        ''' Called when user modified the text of label_in or label_out '''
        for each in self.label_in, self.label_out:
            if not each.hasFocus():
                if not str(each).startswith('['):
                    each.setPlainText('[{}'.format(str(each)))
                if not str(each).endswith(']'):
                    each.setPlainText('{}]'.format(str(each)))
        Signalroute.in_sig = str(self.label_in)[1:-1]
        Signalroute.out_sig = str(self.label_out)[1:-1]


    @property
    def start_point(self):
        ''' Compute connection origin - redefined function '''
        parent_rect = self.parent.boundingRect()
        return QPointF(parent_rect.x(), parent_rect.height() / 2)

    @property
    def end_point(self):
        ''' Compute connection end point - redefined function '''
        # Arrow always bumps at the screen edge
        try:
            view = self.scene().views()[0]
            #view.update_phantom_rect()
            # view_pos is the position of the view relative to the scene
            view_pos = view.mapToScene(
                           view.viewport().geometry()).boundingRect().topLeft()
            scene_pos_x = self.mapFromScene(view_pos).x()
            #print view_pos.x(), scene_pos_x, view.sceneRect().x()
            return QPointF(scene_pos_x, self.start_point.y())
        except (IndexError, AttributeError):
            # In case there is no view (e.g. Export PNG from cmd line)
            return QPointF(self.start_point.x() - 250, self.start_point.y())

    def reshape(self):
        ''' Redefine shape function to add the text areas '''
        super().reshape()

        width_in = self.label_in.boundingRect().width()

        self.label_in.setX(self.start_point.x() - width_in)
        self.label_in.setY(self.start_point.y() + 5)
        self.label_out.setX(self.end_point.x() + 10)
        self.label_out.setY(self.end_point.y() + 5)

    def check_syntax(self, text):
        ''' Check the syntax of the IN and OUT signal lists '''
        try:
            _, syntax_errors, _, _, _ = self.parser.parseSingleElement(
                                           'signalroute', text)
        except (AssertionError, AttributeError) as err:
            LOG.error('Checker failed:' + str(err))
        else:
            return syntax_errors

    def update_completion_list(self, pr_text):
        ''' Called after text has been edited '''
        from .sdlSymbols import CONTEXT
        ast, _, _, _, _ = self.parser.parseSingleElement('signalroute',
                                                         pr_text)
        # ast is a dict:
        # {'routes': [{'dest': str, 'source': str', 'signals': [str]}],
        #  'name': u'c'} - dest and source can be 'ENV'
        all_sigs = []
        for each in ast['routes']:
            source    = each['source']
            dest      = each['dest']
            sigs = [{'name': name} for name in each['signals']]
            all_sigs.extend(sigs)
            def get_sig(signal):
                ''' If signal is declared, return the signature '''
                for sig in CONTEXT.signals:
                    if sig['name'].lower() == signal['name'].lower():
                        return sig
                return signal
            if each['dest'] == 'ENV':
                # output signals of process 'source'
                process, = [p for p in CONTEXT.processes
                            if p.processName.lower() == each['source'].lower()]
                existing = [sig['name'].lower()
                           for sig in process.output_signals]
                for sig in sigs:
                    if sig['name'].lower() not in existing:
                        process.output_signals.append(get_sig(sig))
            else:
                # input signals of process 'source'
                try:
                    process, = [p for p in CONTEXT.processes
                            if p.processName.lower() == each['dest'].lower()]
                    existing = [sig['name'].lower()
                            for sig in process.input_signals]
                    for sig in sigs:
                        if sig['name'].lower() not in existing:
                            process.input_signals.append(get_sig(sig))
                except ValueError:
                    # No process found
                    pass
        existing = [sig['name'].lower() for sig in CONTEXT.signals]
        for each in all_sigs:
            if each['name'].lower() not in existing:
                CONTEXT.signals.append(each)

    def resize_item(self, new_rect):
        ''' Called after signallist text has been edited '''
        pass


class Channel(Signalroute):
    ''' Connection between two processes. Signalroute handles connections
    between a function and the environment (edge of the screen). Here the
    start, middle and end point are redefined. They are stored with
    scene coordinates '''

    @Slot(float, float)
    def child_moved(self, delta_x, delta_y):
        ''' When the connection child moves - redefined function '''
        # compute the distance between the start and end points
        dist_x = abs(self.end_point.x() - self.start_point.x())
        dist_y = abs(self.end_point.y() - self.start_point.y())
        new_dist_x = dist_x - delta_x
        new_dist_y = dist_y - delta_y

        self._end_point.setX(self._end_point.x() - delta_x)
        self._end_point.setY(self._end_point.y() - delta_y)
        x_shift, y_shift = [], []
        middle_points = list(self.middle_points)

        self._middle_points = []
        for ratio, point in zip(self._ratios, middle_points):
            fact_x, fact_y = ratio
            sp = self.start_point
            new_x = (sp.x() + new_dist_x * fact_x) if 0 <= fact_x <= 1 \
                    else point.x() - delta_x
            new_y = (sp.y() + new_dist_y * fact_y) if 0 <= fact_y <= 1 \
                    else point.y() - delta_y
            self._middle_points.append(
                    self.parent.mapToScene(QPointF(new_x, new_y)))

        self.reshape()
        self.update() # force a repaint

    @Slot(float, float)
    def parent_moved(self, delta_x, delta_y):
        ''' When the connection parent moves - redefined function '''
        self.reshape()
        self.update() # force a repaint

    @property
    def start_point(self):
        ''' Compute connection origin - redefined function '''
        return self._start_point

    @start_point.setter
    def start_point(self, scene_coord : QPointF):
        ''' value is in scene coordinates '''
        self._start_point = self.parent.mapFromScene(scene_coord)

    @property
    def end_point(self):
        ''' Compute connection end point - redefined function '''
        return self.parent.mapFromScene(self._end_point)

    @end_point.setter
    def end_point(self, scene_coord : QPointF):
        ''' value is in scene coordinates '''
        self._end_point = scene_coord

    @property
    def middle_points(self):
        ''' Compute connection intermediate points - redefined function '''
        for each in self._middle_points:
            yield self.parent.mapFromScene(each)

    @middle_points.setter
    def middle_points(self, points_scene_coord):
        ''' Redefined function: also store the relative position (percentage)
            to the line length, in order to ensure proper dimensioning of
            the connection when blocks are moved '''
        # compute the distance between the start and end points
        dist_x = abs(self.end_point.x() - self.start_point.x())
        dist_y = abs(self.end_point.y() - self.start_point.y())
        # Compute the distance ratio
        self._ratios = []
        for point in points_scene_coord:
            pCoord = self.parent.mapFromScene(point)
            len_x = abs(pCoord.x() - self.start_point.x())
            len_y = abs(pCoord.y() - self.start_point.y())
            fact_x = 1 if dist_x == 0 else len_x / dist_x
            fact_y = 1 if dist_y == 0 else len_y / dist_y
            if pCoord.y() < self.start_point.y():
                fact_y = -fact_y
            if pCoord.x() < self.start_point.x():
                fact_x = -fact_x
            self._ratios.append((fact_x, fact_y))
        self._middle_points = points_scene_coord

    def add_point(self, scene_coord):
        self._middle_points.append(scene_coord)

    def paint(self, painter, _, ___):
        ''' Apply anti-aliasing '''
        painter.setRenderHint(QPainter.Antialiasing, True)
        super().paint(painter, _, ___)



class Controlpoint(QGraphicsPathItem):
    ''' Class handling one edge control point (to change bezier curves) '''
    user_can_connect = False  # not used but will avoid exceptions
    def __init__(self, pos, edge):
        ''' Set the original control point - with color, shape '''
        path = QPainterPath()
        path.addEllipse(pos.x() - 5, pos.y() - 5, 10, 10)
        super().__init__(path, parent=edge)
        self.setPen(QColor(50, 100, 120, 200))
        self.setBrush(QColor(200, 200, 210, 120))
        self.setFlags(QGraphicsItem.ItemIsSelectable |
                      QGraphicsItem.ItemIsMovable)
        self.edge = edge
        self.hide()

    @property
    def center(self):
        ''' Return the position of the center of the shape '''
        return self.pos() + self.boundingRect().center()

    def mouseMoveEvent(self, event):
        ''' When user moves a control point, update the connection shape '''
        super().mouseMoveEvent(event)
        self.edge.reshape()


class Connectionpoint(Controlpoint):
    '''
        Class handling an end connection point - similar to Controlpoint,
        except the shape and the behaviour when moved. Connection point
        must check with its owning symbol if the move is valid.
    '''
    def __init__(self, pos, edge, symbol):
        ''' Create the point - as a small, lightblue box '''
        super().__init__(pos, edge=edge)
        path = QPainterPath()
        path.addRect(0, 0, 10, 10)
        self.setPath(path)
        self.setPos(pos.x() - 5, pos.y() - 5)
        # Symbol actually owning the connection point
        self.symbol = symbol

    def update_position(self):
        '''
            Update the position of the end point so that it is always
            placed correctly on the symbol that contains it (but that may
            not be the parent of the point if is the end point)
            After the position update, reshape the edge so that the lines
            are properly drawn to the new position
        '''
        # Transform position to owning symbol coordinates:
        center_pos = self.scenePos() + self.boundingRect().center()
        pos_symb = self.symbol.mapFromScene(center_pos)

        # Ask to symbol the nearest valid position on its shape
        # to place the connection point, and adjust position
        nearest_x, nearest_y = self.symbol.closest_connection_point(pos_symb)

        self.moveBy(-nearest_x, -nearest_y)

        self.edge.reshape()

    def mouseMoveEvent(self, event):
        '''
            When user moves the point, check with the connected symbol
            where the actual connection point should be placed
        '''
        # calling super function moves to the new point
        super().mouseMoveEvent(event)
        # Then compute a valid position, based on the owning symbol
        self.update_position()


class Edge(Connection):
    ''' B-spline/Bezier connection shape '''
    def __init__(self, edge, graph):
        ''' Set generic parameters from Connection class '''
        self.text_label = None
        super().__init__(edge['source'], edge['target'])
        self.edge = edge
        self.graph = graph
        # Set connection points as not visible, by default
        self.bezier_visible = False

        # Initialize control point coordinates
        self.bezier = [self.mapFromScene(*self.edge['spline'][0])]
        # Bezier control points (groups of three points):
        assert(len(self.edge['spline']) % 3 == 1)
        for i in range(1, len(self.edge['spline']), 3):
            self.bezier.append([Controlpoint(
                          self.mapFromScene(*self.edge['spline'][i + j]), self)
                          for j in range(3)])

        # Create connection points at start and end of the edge
        self.source_connection = Connectionpoint(
                self.start_point or self.bezier[0], self, self.parent)
        self.parent.movable_points.append(self.source_connection)
        self.end_connection = Connectionpoint(
                self.end_point or self.bezier[-1], self, self.child)
        self.child.movable_points.append(self.end_connection)
        self.reshape()

    @property
    def start_point(self):
        ''' Compute connection origin - redefine in subclasses '''
        # Start point is optional - graphviz decision
        return self.mapFromScene(*self.edge['start']) \
               if self.edge.get('start') else None

    @property
    def end_point(self):
        ''' Compute connection end point - redefine in subclasses '''
        return self.mapFromScene(*self.edge['end']) \
               if self.edge.get('end') else None

    def bezier_set_visible(self, visible=True):
        ''' Display or hide the edge control points '''
        self.bezier_visible = visible
        for group in self.bezier[1:]:
            for ctrl_point in group:
                if visible:
                    ctrl_point.show()
                else:
                    ctrl_point.hide()
        if visible:
            self.end_connection.show()
            self.source_connection.show()
        else:
            self.end_connection.hide()
            self.source_connection.hide()
        self.update()

    def mousePressEvent(self, event):
        ''' On a mouse click, display the control points '''
        self.bezier_set_visible(True)

    # pylint: disable=R0914
    def reshape(self):
        ''' Update the shape of the edge (redefined function) '''
        path = QPainterPath()
        # If there is a starting point, draw a line to the first curve point
        if self.start_point:
            path.moveTo(self.source_connection.center)
            path.lineTo(self.bezier[0])
        else:
            path.moveTo(self.source_connection.center)
        # Loop over the curve points:
        for group in self.bezier[1:]:
            path.cubicTo(*[point.center for point in group])

        # If there is an ending point, draw a line to it
        if self.end_point:
            path.lineTo(self.end_connection.center)

        end_point = path.currentPosition()
        arrowhead = self.angle_arrow(path)
        path.lineTo(arrowhead[0])
        path.moveTo(end_point)
        path.lineTo(arrowhead[1])
        path.moveTo(end_point)
        try:
            # Add the transition label, if any (none for the START edge)
            font = QFont('arial', pointSize=8)
            metrics = QFontMetrics(font)
            label = self.edge.get('label', '') or ''
            lines = label.split('\n')
            longest_line = max(lines, key=len)
            width = metrics.horizontalAdvance(longest_line)
            height = metrics.height() * len(lines)
            # lp is the position of the center of the text
            pos = self.mapFromScene(*self.edge['lp'])
            if not self.text_label:
                self.text_label = QGraphicsTextItem(
                                 self.edge.get('label', ''), parent=self)
                self.text_label.setX(pos.x() - width / 2)
                self.text_label.setY(pos.y() - height / 2)
                self.text_label.setFont(font)
                # Make horizontal center alignment, as dot does
                self.text_label.setTextWidth(
                        self.text_label.boundingRect().width())
                fmt = QTextBlockFormat()
                fmt.setAlignment(Qt.AlignHCenter)
                cursor = self.text_label.textCursor()
                cursor.select(QTextCursor.Document)
                cursor.mergeBlockFormat(fmt)
                cursor.clearSelection()
                self.text_label.setTextCursor(cursor)
                self.text_label.show()
        except KeyError:
            # no label
            pass
        self.setPath(path)

    def __str__(self):
        ''' user-friendly information about the edge coordinates '''
        return('Edge between ' + self.edge['source'].name + ' and ' +
                self.edge['target'].name + str(self.edge['spline'][0]))

    def paint(self, painter, option, widget):
        ''' Apply anti-aliasing to Edge Connections '''
        painter.setRenderHint(QPainter.Antialiasing, True)
        super().paint(painter, option, widget)
        # Draw lines between connection points, if visible
        if self.bezier_visible:
            painter.setPen(
                    QPen(Qt.lightGray, 0, Qt.SolidLine))
            painter.setBrush(Qt.NoBrush)
            points_flat = [point.center
                           for sub1 in self.bezier[1:] for point in sub1]
            painter.drawPolyline([self.source_connection.center]
                                  + points_flat + [self.end_connection.center])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    OpenGEODE - A tiny SDL Editor for TASTE

    This module provides helper functions typically used by backends:

        flatten(ast) : transform a model with nested states to a flat model
        rename_everything(ast, from_name, to_name) : rename symbols
        inner_labels_to_floating(process) : remove labels from transitions
        map_input_state(process) -> mapping: create a mapping
                                    input-state-transition
        sorted_fields(SEQ/CHOICE) : returns the ordered list of fields
                                    of an ASN.1 SEQUENCE or CHOICE type
        state_aggregations: enrich AST with state aggregation flags,
                            and return the list of substates of aggregations
        parallel_states: return a list of strings naming all parallel states
        statenames: return a list of properly-formatted state names
        rec_findstates: recursively find parallel/composite statenames
        update_full_statelist(process): set field full_statelist in process AST
        code_generation_preprocessing: to be called before generating code
        generate_asn1_datamodel: generate the _datamodel.asn file for a process

    Copyright (c) 2012-2022 European Space Agency

    Designed and implemented by Maxime Perrotin

    Contact: maxime.perrotin@esa.int
"""

import operator
import logging
from itertools import chain
from collections import defaultdict

from functools import singledispatch

from . import ogAST

LOG = logging.getLogger(__name__)
DEFAULT_SEPARATOR='_0_'
ASN1SCC = 'asn1Scc'

__all__ = ['flatten', 'rename_everything', 'inner_labels_to_floating',
           'map_input_state', 'sorted_fields', 'state_aggregations',
           'parallel_states', 'statenames', 'rec_findstates',
           'generate_asn1_datamodel',
           'get_full_statelist', 'code_generation_preprocessing']


def statenames(context, sep=DEFAULT_SEPARATOR):
    ''' Return the list of states (just the names) of a given context
    Format the output by replacing unicode separator symbol with a dot '''
    # note: if model has been flattened, all contexts are already merged
    return (s.replace(sep, '.') for s in context.mapping.keys()
            if not s.endswith('START'))


def rec_findstates(context, prefix=''):
    ''' In case of state compositions/aggregations, find substates '''
    for each in context.composite_states:
        if not isinstance(each, ogAST.StateAggregation):
            # Aggregations are just containers, not states
            for name in statenames(each):
                yield f'{prefix}{each.statename}.{name}'
        for substate in rec_findstates(each,
                                       prefix=f'{prefix}{each.statename}.'):
            yield substate


def state_aggregations(process):
    ''' Explore recursively the AST to find all state aggregations, and
        return the composite states inside them
        input: ogAST.Process element
        output: {state_aggregation: {list of ogAST.CompositeState}
    '''
    # { aggregate_name : [list of parallel states] }
    aggregates = defaultdict(list)
    def do_composite(comp, aggregate=''):
        ''' Recursively find all state aggregations in order to allow code
        generator backends to store the state of each parallel state '''
        pre = comp.statename if isinstance(comp, ogAST.StateAggregation) \
                    else ''
        for each in comp.composite_states:
            do_composite(each, pre)
            if isinstance(each, ogAST.StateAggregation):
                # substate of the current state is a state aggregation
                # -> set a flag in each terminator of the current state
                for term in comp.terminators:
                    if term.inputString.lower() == each.statename.lower():
                        term.next_is_aggregation = True
        if aggregate and not isinstance(comp, ogAST.StateAggregation):
            # Composite state inside a state aggregation
            aggregates[aggregate].append(comp)
            # Here, all the terminators inside the composite states must
            # be flagged with the name of the substate so that the NEXTSTATE
            # will not be using the main "context.state" variable but will
            # use the parallel substate name when generating code.
            for each in comp.terminators:
                each.substate = comp.statename
    for each in process.composite_states:
        do_composite(each)
    for each in process.terminators:
        if each.inputString.lower() in aggregates:
            each.next_is_aggregation = True
    for name, comp in aggregates.items():
        # for each state aggregation. update the terminators
        # of each parallel state with the name of all sibling states
        # useful for backends to handle parallel state termination (return)
        # since they have to synchronize with the sibling states
        siblings = [sib.statename for sib in comp]
        for term in [terms for sib in comp for terms in sib.terminators]:
            term.siblings = siblings
    return aggregates


def parallel_states(aggregates):
    ''' Given a mapping obtained with state_aggregation(process), extract
        all parallel states and return a list of state names '''
    parallel_states = []
    for name, comp in aggregates.items():
        for each in comp:
            parallel_states.extend(name for name in each.mapping.keys()
                    if not name.endswith(u'START'))
    return parallel_states


def map_input_state(process):
    ''' Create a mapping dict {input1: {state1: transition, ...}, ...} '''
    mapping = defaultdict(dict)
    input_signals = [sig['name'] for sig in process.input_signals]
    # Add timers to the mapping
    input_signals.extend(process.timers)
    for input_signal in input_signals:
        for state_name, input_symbols in process.mapping.items():
            if isinstance(input_symbols, list):
                # Start symbols have no list of inputs
                for i in input_symbols:
                    if input_signal.lower() in (inp.lower() for
                                               inp in i.inputlist):
                        mapping[input_signal][state_name] = i
    return mapping


def update_full_statelist(process, SEPARATOR=DEFAULT_SEPARATOR) -> None:
    ''' Compute the value of the AST field "process.full_statelist"
        Inputs:
          * process is of type ogAST.Process
    '''

    process.full_statelist = set(chain(process.aggregates.keys(),
                               (name for name in process.mapping.keys()
                                    if not name.endswith('START'))))
    if process.aggregates:
        # Parallel states in a state aggregation may terminate
        process.full_statelist.add(f'state{SEPARATOR}end')

def inner_labels_to_floating(process):
    '''
        Transform inner labels of both floating labels and transitions
        into new floating labels, so that they can be generated in a separate
        section of code, where they are in the scope of everybody
        Works with processes, procedures and nested states
    '''
    for idx in range(len(process.content.floating_labels)):
        for new_floating in find_labels(
                      process.content.floating_labels[idx].transition):
            process.content.floating_labels.append(new_floating)
    for proc_tr in process.transitions:
        for new_floating in find_labels(proc_tr):
            process.content.floating_labels.append(new_floating)
    if process.content.start:
        for new_floating in find_labels(process.content.start.transition):
            process.content.floating_labels.append(new_floating)
    for each in process.content.named_start:
        for new_floating in find_labels(each.transition):
            process.content.floating_labels.append(new_floating)


def flatten(process, sep=u'_'):
    ''' In-place update of the AST: flatten a model with nested states
        Rename inner states, procedures, etc. and move them to process level
    '''
    def update_terminator(context, term, process):
        '''Set next_id, identifying the next transition to run '''
        if term.inputString.lower() in (st.statename.lower()
                                for st in context.composite_states):
            if not term.via:
                term.next_id = term.inputString.lower() + sep + 'START'
            else:
                term.next_id = f'{term.inputString}{sep}{term.entrypoint}_START'
        elif term.inputString.strip() in ('-', '-*'):
            for each in term.possible_states:
                term.candidate_id[-1].append(each)
                for comp in context.composite_states:
                    if each.lower() == comp.statename.lower():
                        if isinstance(comp, ogAST.StateAggregation):
                            term.next_is_aggregation = True
                            term.candidate_id[each + sep + 'START'] = [each]
                        else:
                            term.candidate_id[each + sep + 'START'] = \
                                       [st for st in process.mapping.keys()
                                        if st.startswith(each)
                                        and not st.endswith('START')]
                        continue

    def update_composite_state(state, process):
        ''' Rename inner states, recursively, and add inner transitions
            to process, updating indexes, and update terminators
        '''
        trans_idx = len(process.transitions)
        prefix = state.statename + sep
        set_terminator_states(state, prefix)
        set_transition_states(state, prefix)

        state.mapping = {prefix + key: state.mapping.pop(key)
                         for key in list(state.mapping.keys())}
        # Continuous signal mappings
        state.cs_mapping = {prefix + key: state.cs_mapping.pop(key)
                            for key in list(state.cs_mapping.keys())}

        process.transitions.extend(state.transitions)

        # Add prefix to local variable names and push them at process level
        for dcl in state.variables.keys():
            rename_everything(state.content, dcl, prefix + dcl)
        state.variables = {prefix + key: state.variables.pop(key)
                           for key in list(state.variables.keys())}
        process.variables.update(state.variables)

        # Update return transition indices
        for each in state.terminators:
            if each.kind == 'return':
                for idx, trans in enumerate(process.transitions):
                    if trans == each.next_trans:
                        each.next_id = idx
                        break

        values = []
        for key, value in state.mapping.items():
            # Update transition indices
            if isinstance(value, int):
                # START transitions
                state.mapping[key] = value + trans_idx
            else:
                values.extend(value)

        for each in state.cs_mapping.values():
            # Update transition indices of continuous signals
            # XXX shouldn't we do it also for CONNECT parts?
            values.extend(each)

        for inp in set(values):
            # values may contain duplicate entries if an input corresponds
            # to multiple states. In that case we must update the index of the
            # input only once, thus the set().
            inp.transition_id += trans_idx

        process.mapping.update(state.mapping)
        process.cs_mapping.update(state.cs_mapping)

        # If composite state has entry procedures, add the call
        if state.entry_procedure:
            for each in (trans for trans in state.mapping.values()
                         if isinstance(trans, int)):
                call_entry = ogAST.ProcedureCall()
                call_entry.inputString = 'entry'
                entryproc = f'{prefix}entry'
                call_entry.output = [{'outputName': entryproc,
                                     'params': [], 'tmpVars': []}]
                process.transitions[each].actions.insert(0, call_entry)

        # If composite state has exit procedure, add an call to this
        # procedure if the transition ends up exiting the state with
        # a return statement. There are other calls to the exit procedure
        # that the code generation backend must add when the state is exited
        # from a transition trigger in the super state. See AdaGenerator.py
        if state.exit_procedure:
            # Build up a list of transitions that contain a return statement
            trans_with_return = []
            for each in chain(state.transitions, (lab.transition for lab in
                                              state.content.floating_labels)):
                def rec_transition(trans : ogAST.Transition):
                    if trans.terminator:
                        if trans.terminator.kind == 'return':
                            trans_with_return.append (trans)
                    elif isinstance(trans.actions[-1], ogAST.Decision):
                        # There is no terminator, so the transition may finish
                        # with a DECISION, we must check it recursively
                        for answer in trans.actions[-1].answers:
                            rec_transition (answer.transition)

                rec_transition (each)

            for trans in trans_with_return:
                call_exit = ogAST.ProcedureCall()
                call_exit.inputString = 'exit'
                exitproc = f'{prefix}exit'
                call_exit.output = [{'outputName': exitproc,
                                     'params': [], 'tmpVars': []}]
                trans.actions.append(call_exit)

        for inner in state.composite_states:
            # Go recursively in inner composite states
            inner.statename = prefix + inner.statename
            update_composite_state(inner, process)
            # Remove: recursion is already handled within propagate_inputs
            #propagate_inputs(inner, process)
            #del process.mapping[inner.statename]
        for each in state.terminators:
            # Give prefix to terminators
            if each.label:
                each.label.inputString = prefix + each.label.inputString
            if each.kind == 'next_state':
                if each.inputString.strip() not in ('-', '-*'):
                    each.inputString = prefix + each.inputString
                # Set next transition id
                update_terminator(context=state, term=each, process=process)
            elif each.kind == 'join':
                rename_everything(state.content,
                                  each.inputString,
                                  prefix + each.inputString)
        for each in state.labels:
            # Give prefix to labels in transitions
            rename_everything(state.content,
                              each.inputString,
                              prefix + each.inputString)
        # Add prefixed floating labels of the composite state to the process
        process.content.floating_labels.extend(state.content.floating_labels)
        # Rename inner procedures
        for each in state.content.inner_procedures:
            rename_everything(state.content, each.inputString,
                              prefix + each.inputString)
            each.inputString = prefix + each.inputString
        process.content.inner_procedures.extend(state.content.inner_procedures)

    def propagate_inputs(nested_state, context):
        ''' Nested states: Inputs at level N must be handled at level N-1
            that is, all inputs of a composite states (the ones that allow
            to exit the composite state from the outer scope) must be
            processed by each of the substates.
        '''
        if not isinstance(nested_state, ogAST.StateAggregation):
            for val in nested_state.mapping.values():
                try:
                    inputlist = context.mapping[nested_state.statename]
                    val.extend(inputlist)
                except (AttributeError, KeyError):
                    # KeyError in case of StateAggregation
                    pass
            for val in nested_state.cs_mapping.values():
                try:
                    inputlist = context.cs_mapping[nested_state.statename]
                    val.extend(inputlist)
                except (AttributeError, KeyError):
                    # KeyError in case of StateAggregation
                    pass
        for each in nested_state.composite_states:
            # do the same recursively
            propagate_inputs(each, nested_state)
            #del nested_state.mapping[each.statename]
        if not isinstance(nested_state, ogAST.StateAggregation):
            del context.mapping[nested_state.statename]
            del context.cs_mapping[nested_state.statename]

    def set_terminator_states(context, prefix=''):
        ''' Associate state to terminators, needed to process properly
            a history nextstates (dash nextstate) in code generators '''
        for each in context.content.states:
            for inp in chain(each.inputs,
                             each.continuous_signals,
                             each.connects):
                for term in inp.terminators:
                    term.possible_states.extend(prefix + name.lower()
                                                for name in each.statelist)

    def set_transition_states(context, prefix=''):
        ''' Associate state to transitions, needed to process properly
            the call to the exit procedure of a nested state '''
        for each in context.content.states:
            for inp in chain(each.inputs,
                             each.continuous_signals,
                             each.connects):
                if inp.transition is not None:
                    inp.transition.possible_states.extend(prefix + name.lower()
                                                for name in each.statelist)

    set_terminator_states(process)
    set_transition_states(process)

    for each in process.composite_states:
        update_composite_state(each, process)
        propagate_inputs(each, process)
        #del process.mapping[each.statename]

    # Update terminators at process level
    for each in process.terminators:
        if each.kind == 'next_state':
            update_terminator(process, each, process)


@singledispatch
def rename_everything(ast, from_name, to_name):
    '''
        Rename in all symbols all occurrences of name_ref into new_name.
        This is used to avoid name clashes in nested diagrams when they get
        flattened. For example rename all accesses to a variable declared
        in the scope of a composite state, so that they do not overwrite
        a variable with the same name declared at a higher scope.
    '''
    LOG.debug ('rename_everything - ' + str(ast) + " - ")
    try:
        LOG.debug(ast.inputString)
    except:
        pass

    _, _, _ = ast, from_name, to_name


@rename_everything.register(ogAST.Automaton)
def _rename_automaton(ast, from_name, to_name):
    ''' Renaming at Automaton top level (content of diagrams) '''
    if ast.start:
        rename_everything(ast.start.transition, from_name, to_name)
    for each in ast.named_start:
        rename_everything(each.transition, from_name, to_name)
    for each in ast.floating_labels:
        rename_everything(each, from_name, to_name)
    for each in ast.states:
        for inp in chain(each.inputs, each.continuous_signals, each.connects):
            rename_everything(inp.transition, from_name, to_name)
            for idx, param in enumerate(inp.parameters):
                if param.lower() == from_name.lower():
                    inp.parameters[idx] = to_name
            try:
                rename_everything(inp.trigger, from_name, to_name)
            except AttributeError:
                # only for continuous signals
                pass
        if each.composite:
            rename_everything(each.composite.content, from_name, to_name)
    for each in ast.inner_procedures:
        # Check that from_name is not a redefined variable in the procedure
        for varname in each.variables.keys():
            if varname.lower() == from_name:
                break
        else:
            # do the same for procedure input/output variables (FPAR section)
            for fpar in each.fpar:
                if fpar['name'].lower() == from_name:
                    break
            else:
                rename_everything(each.content, from_name, to_name)


@rename_everything.register(ogAST.Output)
@rename_everything.register(ogAST.ProcedureCall)
def _rename_output(ast, from_name, to_name):
    ''' Rename actual parameter names in output/procedure calls
        and possibly the procedure name '''
    for each in ast.output:
        if each['outputName'].lower() == from_name.lower():
            each['outputName'] = to_name
        for param in each['params']:
            rename_everything(param, from_name, to_name)


@rename_everything.register(ogAST.Label)
def _rename_label(ast, from_name, to_name):
    ''' Rename elements in the transition following a label '''
    if ast.inputString.lower() == from_name.lower():
        ast.inputString = to_name
    rename_everything(ast.transition, from_name, to_name)


@rename_everything.register(ogAST.Decision)
def _rename_decision(ast, from_name, to_name):
    ''' Rename elements in decision '''
    if ast.kind == 'question':
        rename_everything(ast.question, from_name, to_name)
    for each in ast.answers:
        rename_everything(each, from_name, to_name)


@rename_everything.register(ogAST.Answer)
def _rename_answer(ast, from_name, to_name):
    ''' Rename elements in an answer branch '''
    for each in ast.answers:
        if each['kind'] in ('constant', 'open_range'):
            _, constant = each['content'] # get the constant
            rename_everything(constant, from_name, to_name)
        elif each['kind'] == 'closed_range':
            pass  # TODO when supported
    rename_everything(ast.transition, from_name, to_name)


@rename_everything.register(ogAST.Transition)
def _rename_transition(ast, from_name, to_name):
    ''' Rename in all symbols of a transition '''
    for each in chain(ast.actions, ast.terminators):
        # Label, output, task, decision, terminators
        rename_everything(each, from_name, to_name)


@rename_everything.register(ogAST.Terminator)
def _rename_terminator(ast, from_name, to_name):
    ''' Rename terminators: join/labels, next_state '''
    if ast.inputString.lower() == from_name.lower():
        ast.inputString = to_name
    rename_everything(ast.label, from_name, to_name)
    rename_everything(ast.return_expr, from_name, to_name)


@rename_everything.register(ogAST.TaskAssign)
def _rename_task_assign(ast, from_name, to_name):
    ''' List of assignments '''
    for each in ast.elems:
        rename_everything(each, from_name, to_name)


@rename_everything.register(ogAST.TaskForLoop)
def _rename_forloop(ast, from_name, to_name):
    ''' List of FOR loops '''
    for each in ast.elems:
        rename_everything(each['list'], from_name, to_name)
        if each['range'] is not None:
            rename_everything(each['range']['start'], from_name, to_name)
            rename_everything(each['range']['stop'], from_name, to_name)
        if each['transition'] is not None:
            rename_everything(each['transition'], from_name, to_name)


@rename_everything.register(ogAST.ExprPlus)
@rename_everything.register(ogAST.ExprMul)
@rename_everything.register(ogAST.ExprMinus)
@rename_everything.register(ogAST.ExprEq)
@rename_everything.register(ogAST.ExprNeq)
@rename_everything.register(ogAST.ExprGt)
@rename_everything.register(ogAST.ExprGe)
@rename_everything.register(ogAST.ExprLt)
@rename_everything.register(ogAST.ExprLe)
@rename_everything.register(ogAST.ExprDiv)
@rename_everything.register(ogAST.ExprMod)
@rename_everything.register(ogAST.ExprRem)
@rename_everything.register(ogAST.ExprAssign)
@rename_everything.register(ogAST.ExprOr)
@rename_everything.register(ogAST.ExprIn)
@rename_everything.register(ogAST.ExprAnd)
@rename_everything.register(ogAST.ExprXor)
@rename_everything.register(ogAST.ExprAppend)
@rename_everything.register(ogAST.ExprAssign)
def _rename_expr(ast, from_name, to_name):
    ''' Two-sided expressions '''
    rename_everything(ast.left, from_name, to_name)
    rename_everything(ast.right, from_name, to_name)


@rename_everything.register(ogAST.PrimSequenceOf)
def _rename_prim_seq_of(ast, from_name, to_name):
    ''' List of primary '''
    for each in ast.value:
        rename_everything(each, from_name, to_name)

@rename_everything.register(ogAST.PrimSequence)
def _rename_prim_seq(ast, from_name, to_name):
    ''' Values in the fields of a SEQUENCE '''
    for each in ast.value.values():
        rename_everything(each, from_name, to_name)

@rename_everything.register(ogAST.PrimChoiceItem)
def _rename_prim_choice_item(ast, from_name, to_name):
    ''' Value of a CHOICE item '''
    rename_everything(ast.value['value'], from_name, to_name)


@rename_everything.register(ogAST.PrimIndex)
def _rename_index(ast, from_name, to_name):
    ''' Index of an array '''
    rename_everything(ast.value[0], from_name, to_name)
    for each in ast.value[1]['index']:
        rename_everything(each, from_name, to_name)


@rename_everything.register(ogAST.PrimSubstring)
def _rename_substring(ast, from_name, to_name):
    ''' Substrings '''
    rename_everything(ast.value[0], from_name, to_name)
    for each in ast.value[1]['substring']:
        rename_everything(each, from_name, to_name)


@rename_everything.register(ogAST.PrimSelector)
def _rename_primselector(ast, from_name, to_name):
    ''' Rename variable name in field access (a.b.c) '''
    rename_everything(ast.value[0], from_name, to_name)


@rename_everything.register(ogAST.PrimVariable)
def _rename_path(ast, from_name, to_name):
    ''' Ultimate seek point for the renaming: primary path/variables '''
    if ast.value[0].lower() == from_name.lower():
        ast.value[0] = to_name


@rename_everything.register(ogAST.PrimCall)
def _rename_primcall(ast, from_name, to_name):
    ''' PrimCall is used e.g. by function "length" (special operators) '''
    for each in ast.value[1]['procParams']:
        rename_everything(each, from_name, to_name)


@rename_everything.register(ogAST.PrimConditional)
def _rename_ifthenelse(ast, from_name, to_name):
    ''' Rename expressions in Conditional expression construct '''
    for expr in ('if', 'then', 'else'):
        rename_everything(ast.value[expr], from_name, to_name)


def find_labels(trans):
    '''
        Yield a list of transition actions whenever a label is found
        Used to transform labels into floating labels so that the gotos
        in a backend can be contained within a single scope.
    '''
    if not trans:
        return
    # Terminators can have a label - add it to the transition actions
    # (to trigger a goto at code generation)
    if trans.terminator and trans.terminator.label:
        trans.actions.append(trans.terminator.label)
        trans.terminator.label = None
    # Then for each action, check if there are labels and yield
    # a new transition with the remaining actions (following the label)
    for idx, action in enumerate(trans.actions):
        if isinstance(action, ogAST.Label):
            new_trans = ogAST.Transition()
            # Create a floating label
            flab = ogAST.Floating_label(label=action)
            new_trans.actions = \
                    trans.actions[slice(idx + 1, len(trans.actions))]
            new_trans.terminator = trans.terminator
            new_trans.terminators = trans.terminators
            flab.transition = new_trans
            # Transform the label into a JOIN in the original transition
            trans.actions[idx:] = []
            trans.terminator = ogAST.Terminator()
            trans.terminator.inputString = action.inputString
            trans.terminator.kind = 'join'
            # Recursively find labels in the new transition
            for flabel in find_labels(flab.transition):
                yield flabel
            # Then yield the new transition
            yield flab
        elif isinstance(action, ogAST.Decision):
            for answer in action.answers:
                for new_fl in find_labels(answer.transition):
                    # Append the remaining actions of the transition
                    if not new_fl.transition.terminator:
                        new_fl.transition.actions.extend(trans.actions
                                          [slice(idx + 1, len(trans.actions))])
                        new_fl.transition.terminator = trans.terminator
                    yield new_fl


def sorted_fields(atype):
    ''' Return the sorted list of a SEQUENCE or CHOICE type fields '''
    if atype.kind not in ('SequenceType', 'ChoiceType'):
        raise TypeError('Not a SEQUENCE nor a CHOICE')
    tmp = ([k, val.Line, val.CharPositionInLine]
             for k, val in atype.Children.items())
    return (x[0] for x in sorted(tmp, key=operator.itemgetter(1,2)))


def find_basic_type(TYPES, a_type):
    ''' Return the ASN.1 basic type of a_type.
    TYPES is process.dataview
    '''
    basic_type = a_type
    while basic_type.kind == 'ReferenceType':
        # Find type with proper case in the data view
        for typename in TYPES.keys():
            if typename.lower() == basic_type.ReferencedTypeName.lower():
                basic_type = TYPES[typename].type
                break
    return basic_type


def generate_asn1_datamodel(process: ogAST.Process, SEPARATOR: str=DEFAULT_SEPARATOR) -> None:
    ''' Generate an ASN.1 model containing:
          - the state definition
          - a type describing the SDL context
          - all user-defined SDL type (NEWTYPEs...)
        This function is independent from specific backends.
    '''

    if not process.name:
        process.name = process.processName

    process_asn1 = process.name.capitalize().replace('_', '-')
    asn1_template = [
            f'{process_asn1}-Datamodel DEFINITIONS ::=',
            'BEGIN',
            '--  This file was generated automatically by OpenGEODE']

    # The ASN.1 module must import the types from other asn1 modules
    types_with_proper_case, imports = [], []
    imports = []
    for moduleName, sorts in process.DV.exportedTypes.items():
        actual_sorts = []
        for each in sorts:
            process.mapping_sort_module[each] = moduleName
            if process.DV.types[each].AddedType == "False":
                actual_sorts.append(each)
                types_with_proper_case.append(each)
        sortsAsList = ", ".join(actual_sorts)
        if sortsAsList:
            imports.append(f'{sortsAsList} FROM {moduleName}')
    if types_with_proper_case:
        asn1_template.append(
                "IMPORTS\n   " + "\n   ".join(imports) + ";")

    # Format the state list with ASN.1-compatible syntax
    statelist_asn1 = (state.lower().replace('_', '-')
                      for state in process.full_statelist)
    states_asn1 = ", ".join(statelist_asn1) \
            or f'{process_asn1.lower()}-has-no-state'
    # Create an ASN.1 definition of the list of states, instead of
    # a native Ada type - this allows external tools to access
    # information about the model without having to parse SDL
    asn1_states_def = f"\n{process_asn1}-States ::= ENUMERATED" \
                      f" {{{states_asn1}}}"

    asn1_template.append(asn1_states_def)

    # Template to generate the full SDL context as a single ASN.1 type
    # This is very useful then to manipulate it from C, Ada or Python as
    # it is then directly accessible and uPER-encodable
    context_elems = []
    if process.full_statelist:
        # Some systems may have no states - only a start transition
        context_elems.append(f'   state {process_asn1}-States')

    context_elems.append ('init-done BOOLEAN')
    # State aggregation: add list of substates
    for substates in process.aggregates.values():
        for each in substates:
            context_elems.append(
              f'{each.statename.lower()}{SEPARATOR}state {process_asn1}-States')

    for var_name, (var_type, def_value) in process.variables.items():
        if var_name in process.aliases.keys():
            continue
        try:
            sortname = var_type.ReferencedTypeName
        except AttributeError:
            # if the DCL variable references a non-existing type, it shall
            # not prevent the model saving.
            continue
        # If the type is a ends with _selection, i.e. it is an OG-created
        # CHOICE selector, then prefix the type with the process name
        if sortname.endswith('-selection'):
            sortname = f'{process.name}-{sortname}'.title()
        context_elems.append(f'{var_name.lower()} {sortname}')

    asn1_context = (f'\n{process_asn1}-Context ::='' SEQUENCE {\n'
                    + ",\n   ".join(line.replace("_", "-").replace("'", '"') for line in context_elems)
                    + "\n}\n")

    asn1_template.append(asn1_context)

    # Add user-defined NEWTYPEs and CHOICE selector types
    choice_selections = []
    for sortname, sortdef in process.user_defined_types.items():
        if sortdef.type.kind == "SequenceOfType":
            rangeMin = sortdef.type.Min
            rangeMax = sortdef.type.Max
            refType  = sortdef.type.type.ReferencedTypeName
            for refTypeCase in types_with_proper_case:
                if refTypeCase.lower().replace('-', '_') == \
                        refType.lower().replace('-', '_'):
                    break
            asn1_template.append(
                    f'{sortname.replace("_", "-").capitalize()} ::= SEQUENCE '
                    f'(SIZE ({rangeMin} .. {rangeMax})) OF '
                    f'{refTypeCase.replace("_", "-")}')
        elif sortdef.type.kind == "EnumeratedType":
            # At the moment, the only user-defined ENUMERATED types that are
            # supported are the ones generated for the CHOICE selectors
            # We can therefore safely rename them systematically here,
            # by using the process name as prefix, to avoid any risk of
            # duplicate type definition in case it exists in another SDL
            # process of the system.
            prefixed_name = f'{process.name}_{sortdef.ChoiceTypeName}_selection '
            keys = []
            for idx, key in enumerate(sortdef.type.EnumValues.keys()):
                # give an index to the enumerations to align with -selection
                # types used in choice index
                keys.append(f'{key}-present({idx+1})')
            asn1_template.append(
                    f'{prefixed_name.replace("_", "-").title()} ::= ENUMERATED {{' + ", ".join(keys) + '}')
    asn1_template.append('END\n')

    # Write the ASN.1 file
    with open(process.name.lower() + '_datamodel.asn', 'w') as asn1_file:
        asn1_file.write('\n'.join(asn1_template))



def code_generation_preprocessing(process, separator=DEFAULT_SEPARATOR):
    ''' Do all sorts of preprocessing before invoking a code generator
        and update the AST of the process
    '''
    # In case model has nested states, flatten everything
    flatten(process, sep=separator)

    # Pre-processing of aliases: we must rename all references to the aliases
    # when they point to a structure that contain a CHOICE field (we cannot
    # use a rename clause in that case, since the field depends on a
    # discriminant.
    no_renames = []
    for (alias, (sort, alias_expr)) in process.aliases.items():
        def rec_detect_choice(expr: ogAST.Expression) -> bool:
            if not isinstance(expr, ogAST.PrimSelector):
                return False
            receiver = expr.value[0]
            bty = find_basic_type(TYPES=process.dataview,
                                  a_type=receiver.exprType)
            if bty.kind == 'ChoiceType':
                return True
            return rec_detect_choice(receiver)
        is_choice = rec_detect_choice(alias_expr)
        if is_choice:
            LOG.debug(f"alias: {alias_expr.inputString} will replace {alias}")
            rename_everything(process.content,
                              alias,
                              alias_expr.inputString)
            no_renames.append(alias)

    # Process State aggregations (Parallel states)

    # Find recursively in the AST all state aggregations
    # Format: {'aggregation_name' : [list of ogAST.CompositeState]
    process.aggregates = state_aggregations(process)

    # Make an mapping {input: {state: transition...}} in order to easily
    # generate the lookup tables for the state machine runtime
    process.input_mapping = map_input_state(process)

    # Extract the list of parallel states names inside the composite states
    # of state aggregations
    process.parallel_states = parallel_states(process.aggregates)

    # Establish the list of states (excluding START states)
    update_full_statelist(process, separator)

    return no_renames
